@@ -1,8 +1,7 @@
-#!/usr/bin/env python
+#!/usr/bin/env python3
 # coding: utf8
 """Python3 link-grammar test script"""
 
-from __future__ import print_function
 import sys, os, re
 import locale
 import unittest
@@ -32,14 +31,14 @@ print(clg.linkgrammar_get_configuration())
 for imported_module in 'linkgrammar$', 'clinkgrammar', '_clinkgrammar', 'lg_testutils':
     module_found = False
     for module in sys.modules:
-        if re.search(r'^(linkgrammar\.)?'+imported_module, module):
+        if re.search(r'^(linkgrammar\.)?' + imported_module, module):
             print("Using", sys.modules[module], end='')
             if hasattr(sys.modules[module], '__version__'):
                 print(' version', sys.modules[module].__version__, end='')
             print()
             module_found = True
     if not module_found:
-        print("Warning: Module", imported_module,  "not loaded.")
+        print("Warning: Module", imported_module, "not loaded.")
 
 sys.stdout.flush()
 #===
@@ -171,7 +170,7 @@ class BParseOptionsTestCase(unittest.TestCase):
 
     def test_setting_verbosity_to_not_allow_value_raises_value_error(self):
         po = ParseOptions()
-        self.assertRaises(ValueError, setattr, po, "verbosity", 121)
+        self.assertRaises(ValueError, setattr, po, "verbosity", -1)
 
     def test_setting_verbosity_to_non_integer_raises_type_error(self):
         po = ParseOptions()
@@ -353,7 +352,7 @@ class DBasicParsingTestCase(unittest.TestCase):
     def test_that_parse_returns_empty_iterator_on_no_linkage_sat(self):
         """Parsing a bad sentence with no null-links shouldn't give any linkage (sat)"""
         self.po = ParseOptions(use_sat=True)
-        if self.po.use_sat != True:
+        if not self.po.use_sat:
             raise unittest.SkipTest("Library not configured with SAT parser")
         result = self.parse_sent("This this doesn't parse", self.po)
         linkage_exists = False
@@ -423,7 +422,7 @@ class DBasicParsingTestCase(unittest.TestCase):
 class EErrorFacilityTestCase(unittest.TestCase):
     # Initialize class variables to invalid (for the test) values.
     handler = {
-        "default":  lambda x, y=None: None,
+        "default": lambda x, y=None: None,
         "previous": lambda x, y=None: None
     }
 
@@ -464,7 +463,7 @@ class EErrorFacilityTestCase(unittest.TestCase):
         self.assertEqual(self.__class__.handler["previous"].__name__, "error_handler_test")
         self.assertRaises(LG_Error, Dictionary, "seh_dummy2")
         self.gotit = None
-        for i in range(0, 2+self.testleaks):
+        for i in range(0, 2 + self.testleaks):
             self.numerr = LG_Error.printall(self.error_handler_test, self)
             if i == 0:
                 self.assertEqual(self.numerr, 1)
@@ -508,13 +507,13 @@ class EErrorFacilityTestCase(unittest.TestCase):
                                 self.errinfo, "bad param")
         self.assertRaisesRegexp(ValueError, "must be an integer",
                                 self.__class__.handler["default"],
-                                self.errinfo, clg.lg_None+1)
+                                self.errinfo, clg.lg_None + 1)
         self.assertRaises(ValueError, self.__class__.handler["default"],
                           self.errinfo, -1)
 
         try:
             self.param_ok = False
-            save_stdout  = divert_start(1) # Note: Handler parameter is stdout
+            save_stdout = divert_start(1) # Note: Handler parameter is stdout
             self.__class__.handler["default"](self.errinfo, 1)
             self.assertIn(dummy_lang, save_stdout.divert_end())
             self.param_ok = True
@@ -549,7 +548,7 @@ class EErrorFacilityTestCase(unittest.TestCase):
 
     def test_30_formatmsg(self):
         # Here the error handler is still set to None.
-        for _ in range (0, 1+self.testleaks):
+        for _ in range(0, 1 + self.testleaks):
             self.assertRaises(LG_Error, Dictionary, "formatmsg-test-dummy-dict")
             LG_Error.printall(self.error_handler_test, self)
             self.assertRegexpMatches(self.errinfo.formatmsg(), "link-grammar: Error: .*formatmsg-test-dummy-dict")
@@ -584,10 +583,9 @@ class EErrorFacilityTestCase(unittest.TestCase):
 
     def test_50_set_orig_error_handler(self):
         # Set the error handler back to the default handler.
-        # The error message is now visible (but we cannot test that).
         self.__class__.handler["previous"] = LG_Error.set_handler(self.__class__.handler["default"])
         self.assertIsNone(self.__class__.handler["previous"])
-        for _ in range(0, 1+self.testleaks):
+        for _ in range(0, 1 + self.testleaks):
             self.__class__.handler["previous"] = LG_Error.set_handler(self.__class__.handler["default"])
         self.assertEqual(self.__class__.handler["previous"].__name__, "_default_handler")
 
@@ -602,7 +600,7 @@ class FSATsolverTestCase(unittest.TestCase):
     def setUp(self):
         self.d, self.po = Dictionary(lang='en'), ParseOptions()
         self.po = ParseOptions(use_sat=True)
-        if self.po.use_sat != True:
+        if not self.po.use_sat:
             raise unittest.SkipTest("Library not configured with SAT parser")
 
     def test_SAT_getting_links(self):
@@ -673,7 +671,7 @@ class HEnglishLinkageTestCase(unittest.TestCase):
     # -- the one that is in the dict is not the grammatically appropriate word.
     #
     # Let's is NOT split into two! It's in the dict as one word, lower-case only.
-    def test_f_captilization(self):
+    def test_f_capitalization(self):
         self.assertEqual(list(self.parse_sent('Let\'s eat.')[0].words()),
              ['LEFT-WALL', 'let\'s', 'eat.v', '.', 'RIGHT-WALL'])
 
@@ -779,7 +777,6 @@ class HEnglishLinkageTestCase(unittest.TestCase):
 "\nLEFT-WALL I.p enjoy.v eating.v bass.n-u ."
 "\n\n")
 
-
         sent = 'We are from the planet Gorpon'
         linkage = self.parse_sent(sent)[0]
         self.assertEqual(linkage.diagram(),
@@ -824,11 +821,12 @@ class IWordPositionTestCase(unittest.TestCase):
     def test_en_word_positions(self):
         linkage_testfile(self, self.d_en, ParseOptions(), 'pos')
 
-    def test_en_spell_word_positions(self):
-        po = ParseOptions(spell_guess=1)
-        if po.spell_guess == 0:
-            raise unittest.SkipTest("Library is not configured with spell guess")
-        linkage_testfile(self, self.d_en, po, 'pos-spell')
+# Temporarily stub out spell checking. See issue #1249 for status.
+#    def test_en_spell_word_positions(self):
+#        po = ParseOptions(spell_guess=1)
+#        if po.spell_guess == 0:
+#            raise unittest.SkipTest("Library is not configured with spell guess")
+#        linkage_testfile(self, self.d_en, po, 'pos-spell')
 
     def test_ru_word_positions(self):
         linkage_testfile(self, Dictionary(lang='ru'), ParseOptions(), 'pos')
@@ -971,6 +969,149 @@ class JBDictCostReadingTestCase(unittest.TestCase):
         linkage = Sentence('Is the bed white?', self.d, self.po).parse().next()
         self.assertEqual(list(linkage.words())[4], 'white.a')
 
+
+def sm(s):
+    """
+    Naive '.' replacement by SUBSCRIPT_MARK (enough for the tests here).
+    """
+    SUBSCRIPT_MARK = '\3'
+    return s.replace('.', SUBSCRIPT_MARK)
+
+
+class XLookupListTestCase(unittest.TestCase):
+    """
+    Lookup-list related tests.
+    Note: In case of test errors, the subscript mark in words may be mis-printed.
+    """
+    @classmethod
+    def setUpClass(cls):
+        cls.d_en, cls.d_sql, cls.po = Dictionary(lang='en'), Dictionary(lang='demo-sql'), ParseOptions()
+
+    @classmethod
+    def tearDownClass(cls):
+        del cls.d_en, cls.d_sql, cls.po
+
+    def test_file_lookup_list_none(self):
+        self.assertIsNone(clg.dictionary_lookup_list(self.d_en._obj, 'NoSuchWord'))
+
+    def test_file_lookup_wild_none(self):
+        self.assertIsNone(clg.dictionary_lookup_wild(self.d_en._obj, 'NoSuch*'))
+
+    def test_sql_lookup_list_none(self):
+        self.assertIsNone(clg.dictionary_lookup_list(self.d_sql._obj, 'NoSuchWord'))
+
+    def test_sql_lookup_wild_none(self):
+        self.assertIsNone(clg.dictionary_lookup_wild(self.d_sql._obj, 'NoSuch*'))
+
+    def test_file_lookup_list_subscr(self):
+        dictnode = clg.dictionary_lookup_list(self.d_en._obj, sm('test.n'))
+        self.assertEqual(dictnode[0].string, sm('test.n'))
+
+    def test_file_lookup_list_no_subscr(self):
+        dictnode = clg.dictionary_lookup_list(self.d_en._obj, 'test')
+        self.assertEqual(sorted([dictnode[i].string for i in range(len(dictnode))]), [sm('test.n'), sm('test.v')])
+        for i in range(len(dictnode)):
+            self.assertIn("(", str(dictnode[i].exp), "Missing expression")
+            if dictnode[i].string == sm('test.n'):
+                self.assertEqual(dictnode[i].file, 'en/words/words.n.1-const')
+            elif dictnode[i].string == sm('test.v'):
+                self.assertIsNone(dictnode[i].file)
+
+    @unittest.skip("FIXME: Cannot lookup with SUBSCRIPT_MARK")
+    def test_sql_lookup_list_subscr(self):
+        dictnode = clg.dictionary_lookup_list(self.d_en._obj, sm('test.n'))
+        self.assertEqual(dictnode[0].string, sm('test.n'))
+
+    @unittest.skip("FIXME: It returns a dot subscript instead of SUBSCRIPT_MARK")
+    def test_sql_lookup_list_no_subscr(self):
+        dictnode = clg.dictionary_lookup_list(self.d_sql._obj, 'test')
+        self.assertEqual(sorted([dictnode[i].string for i in range(len(dictnode))]), [sm('test.n')])
+
+    def test_file_lookup_wild_any_subscript(self):
+        dictnode = clg.dictionary_lookup_wild(self.d_en._obj, 'test*')
+        self.assertTrue(len(dictnode) > 40, 'Missing words (only {} found)'.format(len(dictnode)))
+        for dn in dictnode:
+            self.assertIsNotNone(re.search('test.*', dn.string), 'Bad word {}'.format(dn.string))
+
+    def test_file_lookup_wild_n_subscript(self):
+        dictnode = clg.dictionary_lookup_wild(self.d_en._obj, 'test*.n')
+        self.assertTrue(len(dictnode) > 10, 'Missing words (only {} found)'.format(len(dictnode)))
+        for dn in dictnode:
+            self.assertIsNotNone(re.search('test.*' + sm('.') + 'n', dn.string), 'Bad word {}'.format(dn.string))
+
+    def test_sql_lookup_wild_any_subscript(self):
+        dictnode = clg.dictionary_lookup_wild(self.d_en._obj, 't*')
+        self.assertTrue(len(dictnode) > 40, 'Missing words (only {} found)'.format(len(dictnode)))
+        for dn in dictnode:
+            self.assertIsNotNone(re.search('t.*', dn.string), 'Bad word {}'.format(dn.string))
+
+
+# Noun expressions currently contain only a dialect component "headline".
+# It controls the cost of a null expression in 2 places.
+class XExp_resolving_test(unittest.TestCase):
+    @classmethod
+    def setUpClass(cls):
+        cls.d = Dictionary(lang='en')
+
+    @classmethod
+    def tearDownClass(cls):
+        del cls.d
+
+    def test_exp_copying(self):
+        dictnode = clg.dictionary_lookup_list(self.d._obj, sm('test.n'))
+        exp_old = dictnode[0].exp
+        exp_new = clg.lg_exp_resolve(self.d._obj, exp_old)
+        self.assertEqual(str(exp_old), str(exp_new))
+
+    # When a zero cost is assigned to "headline", the expression should remain the same.
+    def test_no_op_resolving(self):
+        dictnode = clg.dictionary_lookup_list(self.d._obj, sm('test.n'))
+        exp_old = dictnode[0].exp
+        exp_new = clg.lg_exp_resolve(self.d._obj, exp_old, ParseOptions(dialect="headline:0")._obj)
+        self.assertEqual(str(exp_old), str(exp_new))
+
+    def test_resolving(self):
+        """
+        Test expression resolving using the default headline:4 setting from
+        data/en/4.0.dialect.
+        """
+        dictnode = clg.dictionary_lookup_list(self.d._obj, sm('test.n'))
+        exp_old = dictnode[0].exp
+        exp_new = clg.lg_exp_resolve(self.d._obj, exp_old, ParseOptions()._obj)  # headline:4
+
+        # Find the 2 locations with a difference when comparing
+        # exp_old to exp_new and validate them.
+        # Locations 2 and 4 are in exp_old. Locations 6 and 7 are in exp_new.
+        # () is the null expression that is controlled by "headline".
+        str_comb = str(exp_old) + '%' + str(exp_new)
+        diff = re.search(r'^([^%]*)(\(Ds\*\*x- or \([^%]*?\)\)\)\))([^%]*)(\(Ds\*\*c- or \([^%]*?\)\)\)\))([^%]*)%\1(.*)\3(.*)\5$', str_comb)
+        #                      1                2                     3                4                     5        6     7
+        self.assertEqual(diff.group(2), '(Ds**x- or (())))')
+        self.assertEqual(diff.group(4), '(Ds**c- or (())))')
+        self.assertEqual(diff.group(6), '(Ds**x- or ([[[[()]]]])))')
+        self.assertEqual(diff.group(7), '(Ds**c- or ([[[[()]]]])))')
+
+
+# Currently, the dictionary creating function sets the generation mode if
+# the "test" parse-option has "generate" in its value list. So it must be
+# set so before the call to Dictionary(). When the second argument of
+# Sentence is evaluated, it initializes the test parse-option to a null
+# string.
+class YGenerationTestCase(unittest.TestCase):
+    """
+    Generation mode tests.
+    """
+    def test_getting_linkages_file_dict(self):
+        ParseOptions(test='generate')
+        linkages = Sentence((clg.WILDCARD_WORD + ' ') * 5, Dictionary(lang='lt'), ParseOptions()).parse()
+        self.assertTrue(len(linkages) > 0, "No linkages")
+
+    def test_getting_linkages_sql_dict(self):
+        ParseOptions(test='generate')
+        linkages = Sentence((clg.WILDCARD_WORD + ' ') * 4, Dictionary(lang='demo-sql'), ParseOptions()).parse()
+        self.assertTrue(len(linkages) > 0, "No linkages")
+
+
 class ZENConstituentsCase(unittest.TestCase):
     @classmethod
     def setUpClass(cls):
@@ -1192,14 +1333,14 @@ def linkage_testfile(self, lgdict, popt, desc=''):
 
         # Lines starting with "-" contain a Parse Option
         elif line[0] == '-':
-                exec('popt.' + line[1:]) in {}, locals()
+            exec('popt.' + line[1:]) in {}, locals()
 
         elif line[0] in '%\r\n':
             pass
         else:
             self.fail('\nTest file "{}": Invalid opcode "{}" (ord={})'.format(testfile, line[0], ord(line[0])))
 
-    self.assertIn(last_opcode , 'OCP', "Missing result comparison in " + testfile)
+    self.assertIn(last_opcode, 'OCP', "Missing result comparison in " + testfile)
 
 def warning(*msg):
     progname = os.path.basename(sys.argv[0])
@@ -1241,12 +1382,12 @@ lg_testutils.add_eqcost_linkage_order(Sentence)
 # "test", "debug" and "verbosity" options. The following allows to specify them
 # as "tests.py" arguments, interleaved with standard "unittest" arguments.
 
-for i,arg in enumerate(sys.argv):
+for i, arg in enumerate(sys.argv):
     debug = sys.argv.pop(i)[7:] if arg.startswith('-debug' + '=') else ''
-for i,arg in enumerate(sys.argv):
-    test = sys.argv.pop(i)[6:] if arg.startswith('-test' + '=')  else ''
-for i,arg in enumerate(sys.argv):
-    verbosity = int(sys.argv.pop(i)[11:]) if arg.startswith('-verbosity' + '=')  else ''
+for i, arg in enumerate(sys.argv):
+    test = sys.argv.pop(i)[6:] if arg.startswith('-test' + '=') else ''
+for i, arg in enumerate(sys.argv):
+    verbosity = int(sys.argv.pop(i)[11:]) if arg.startswith('-verbosity' + '=') else ''
 if (test or debug or verbosity):
     ParseOptions = lg_testutils.add_test_option(ParseOptions, test, debug, verbosity)
 
