@@ -14,6 +14,7 @@
 #include <string.h>
 
 #include "api-structures.h"
+#include "dict-common/dict-defines.h"
 #include "dict-common/dialect.h"
 #include "dict-common/dict-utils.h"
 #include "disjunct-utils.h"             // free_sentence_disjuncts
@@ -88,15 +89,14 @@ Parse_Options parse_options_create(void)
 
 	/* A cost of 2.7 allows the usual cost-2 connectors, plus the
 	 * assorted fractional costs, without going to cost 3.0, which
-	 * is used only during panic-parsing.
-	 * XXX In the long run, this should be fetched from the dictionary
-	 * (and should probably not be a parse option).
-	 */
-	po->disjunct_cost = 2.7;
+	 * is used only during panic-parsing. */
+	po->disjunct_cost = UNINITIALIZED_MAX_DISJUNCT_COST;
 	po->min_null_count = 0;
 	po->max_null_count = 0;
 	po->islands_ok = false;
+#if USE_SAT_SOLVER
 	po->use_sat_solver = false;
+#endif
 	po->linkage_limit = 100;
 #if defined HAVE_HUNSPELL || defined HAVE_ASPELL
 	po->use_spell_guess = 7;
@@ -198,7 +198,17 @@ void parse_options_set_debug(Parse_Options opts, const char * dummy)
 }
 
 char * parse_options_get_debug(Parse_Options opts) {
-	return opts->debug;
+	static char buff[256]; /* see parse_options_set_debug() */
+	char *buffp = buff;
+
+	/* Remove the added commas. */
+	strcpy(buff, opts->debug);
+	if (buffp[0] == ',')
+		buffp++;
+	if ((buffp[0] != '\0') && buffp[strlen(buffp)-1] == ',')
+		buffp[strlen(buffp)-1] = '\0';
+
+	return buffp;
 }
 
 void parse_options_set_test(Parse_Options opts, const char * dummy)
@@ -235,7 +245,17 @@ void parse_options_set_test(Parse_Options opts, const char * dummy)
 }
 
 char * parse_options_get_test(Parse_Options opts) {
-	return opts->test;
+	static char buff[256]; /* see parse_options_set_test() */
+	char *buffp = buff;
+
+	/* Remove the added commas. */
+	strcpy(buff, opts->test);
+	if (buffp[0] == ',')
+		buffp++;
+	if ((buffp[0] != '\0') && buffp[strlen(buffp)-1] == ',')
+		buffp[strlen(buffp)-1] = '\0';
+
+	return buffp;
 }
 
 void parse_options_set_use_sat_parser(Parse_Options opts, bool dummy) {
@@ -251,7 +271,11 @@ void parse_options_set_use_sat_parser(Parse_Options opts, bool dummy) {
 }
 
 bool parse_options_get_use_sat_parser(Parse_Options opts) {
+#if USE_SAT_SOLVER
 	return opts->use_sat_solver;
+#else
+	return false;
+#endif
 }
 
 void parse_options_set_linkage_limit(Parse_Options opts, int dummy)
@@ -429,12 +453,15 @@ Sentence sentence_create(const char *input_string, Dictionary dict)
 	sent->postprocessor = post_process_new(dict->base_knowledge);
 
 	/* Make a copy of the input */
-	sent->orig_sentence = string_set_add (input_string, sent->string_set);
+	sent->orig_sentence = string_set_add(input_string, sent->string_set);
 
 	/* Set the minimum length for tracon sharing.
 	 * In that case a tracon list is produced for the pruning step,
 	 * and disjunct/connector packing is done too. */
-	sent->min_len_encoding = SENTENCE_MIN_LENGTH_TRAILING_HASH;
+	if (IS_GENERATION(dict))
+		sent->min_len_encoding = 0;
+	else
+		sent->min_len_encoding = SENTENCE_MIN_LENGTH_TRAILING_HASH;
 	const char *min_len_encoding = test_enabled("min-len-encoding");
 	if (NULL != min_len_encoding)
 		sent->min_len_encoding = atoi(min_len_encoding+1);
@@ -505,7 +532,7 @@ static void free_sentence_words(Sentence sent)
 	{
 		free(sent->word[i].alternatives);
 	}
-	free_sentence_disjuncts(sent);
+	free_sentence_disjuncts(sent, /*categories_too*/true);
 	free((void *) sent->word);
 	sent->word = NULL;
 }
@@ -520,19 +547,30 @@ void sentence_delete(Sentence sent)
 	free_linkages(sent);
 	post_process_free(sent->postprocessor);
 	post_process_free(sent->constituent_pp);
-	lg_exp_stringify(NULL);
+	exp_stringify(NULL);
+	free(sent->disjunct_used);
 
 	global_rand_state = sent->rand_state;
-	pool_delete(sent->fm_Match_node);
+	pool_delete(sent->Match_node_pool);
 	pool_delete(sent->Table_connector_pool);
+	pool_delete(sent->wordvec_pool);
 	pool_delete(sent->Exp_pool);
 	pool_delete(sent->X_node_pool);
 	if (IS_DB_DICT(sent->dict))
 	{
+#if 0 /* Cannot reuse in case a previous sentence is not deleted yet. */
 		condesc_reuse(sent->dict);
+#endif
 		pool_reuse(sent->dict->Exp_pool);
 	}
 
+	if (NULL != sent->wildcard_word_dc_memblock)
+	{
+		free_categories_from_disjunct_array(sent->wildcard_word_dc_memblock,
+		                                    sent->wildcard_word_num_disjuncts);
+		free(sent->wildcard_word_dc_memblock);
+	}
+
 	free(sent);
 }
 
@@ -597,6 +635,25 @@ int sentence_parse(Sentence sent, Parse_Options opts)
 {
 	int rc;
 
+	if (IS_GENERATION(sent->dict))
+	{
+#if USE_SAT_SOLVER
+		if (opts->use_sat_solver)
+		{
+			prt_error("Error: Cannot use the SAT parser in generation mode\n");
+			return -3;
+		}
+#endif
+		if (opts->max_null_count > 0)
+		{
+			prt_error("Error: Cannot parse with nulls in generation mode\n");
+			return -3;
+		}
+	}
+
+	if (opts->disjunct_cost == UNINITIALIZED_MAX_DISJUNCT_COST)
+		opts->disjunct_cost = sent->dict->default_max_disjunct_cost;
+
 	sent->num_valid_linkages = 0;
 
 	/* If the sentence has not yet been split, do so now.
@@ -614,7 +671,7 @@ int sentence_parse(Sentence sent, Parse_Options opts)
 		 * garbage. Free it. We really should make the code that is panicking
 		 * do this free, but right now, they have no API for it, so we do it
 		 * as a favor. XXX FIXME someday. */
-		free_sentence_disjuncts(sent);
+		free_sentence_disjuncts(sent, /*categories_too*/true);
 	}
 
 	/* Check for bad sentence length */
@@ -632,22 +689,30 @@ int sentence_parse(Sentence sent, Parse_Options opts)
 	if (IS_DB_DICT(sent->dict))
 		condesc_setup(sent->dict);
 
+	for (WordIdx w = 0; w < sent->length; w++)
+	{
+		for (X_node *x = sent->word[w].x; x != NULL; x = x->next)
+			set_connector_farthest_word(x->exp, (int)w, (int)sent->length, opts);
+	}
+
 	/* Expressions were set up during the tokenize stage.
 	 * Prune them, and then parse.
 	 */
 	expression_prune(sent, opts);
 	print_time(opts, "Finished expression pruning");
+#if USE_SAT_SOLVER
 	if (opts->use_sat_solver)
 	{
 		sat_parse(sent, opts);
 	}
 	else
+#endif
 	{
 		classic_parse(sent, opts);
 	}
 	print_time(opts, "Finished parse");
 
-	if ((verbosity > 0) &&
+	if ((verbosity > 0) && !IS_GENERATION(sent->dict) &&
 	   (PARSE_NUM_OVERFLOW < sent->num_linkages_found))
 	{
 		prt_error("Warning: Combinatorial explosion! nulls=%u cnt=%d\n"
@@ -662,12 +727,30 @@ int sentence_parse(Sentence sent, Parse_Options opts)
  * Definitions for linkgrammar_get_configuration().
  */
 
+#define MACVAL(macro) #macro lg_str(=macro) " "
+
 #ifdef __STDC_VERSION__
-#define LG_S1 "__STDC_VERSION__=" lg_xstr(__STDC_VERSION__)
+#define LG_S1 MACVAL(__STDC_VERSION__)
 #else
 #define LG_S1
 #endif
 
+#ifdef _POSIX_C_SOURCE
+#define LG_S2 MACVAL(_POSIX_C_SOURCE)
+#else
+#define LG_S2
+#endif
+
+#if !defined _POSIX_C_SOURCE || _POSIX_C_SOURCE == 0
+ #ifdef _POSIX_SOURCE
+ #define LG_S3 MACVAL(_POSIX_SOURCE)
+ #else
+ #define LG_S3
+ #endif
+#else
+#define LG_S3
+#endif
+
 /* -DCC=$(CC) is added in the Makefile. */
 #ifdef CC
 #define LG_CC CC
@@ -678,19 +761,19 @@ int sentence_parse(Sentence sent, Parse_Options opts)
 #endif
 
 #ifdef __VERSION__
-#define LG_V1 "__VERSION__=" lg_xstr(__VERSION__)
+#define LG_V1 MACVAL(__VERSION__)
 #else
 #define LG_V1
 #endif
 
 #ifdef _MSC_FULL_VER
-#define LG_V2 "_MSC_FULL_VER=" lg_xstr(_MSC_FULL_VER)
+#define LG_V2 MACVAL(_MSC_FULL_VER)
 #else
 #define LG_V2
 #endif
 
 #define LG_COMP LG_CC " " LG_V1 " " LG_V2
-#define LG_STD LG_S1
+#define LG_STD LG_S1 LG_S2 LG_S3
 
 #ifdef __unix__
 #define LG_unix "__unix__ "
