@@ -16,6 +16,7 @@
 #include "dict-common/dict-affix.h"
 #include "dict-common/dict-api.h"
 #include "dict-common/dict-common.h"
+#include "dict-common/dict-defines.h"
 #include "dict-common/dict-impl.h"
 #include "dict-common/dict-utils.h"
 #include "dict-common/file-utils.h"
@@ -113,6 +114,8 @@ dictionary_six_str(const char * lang,
 	dict = (Dictionary) malloc(sizeof(struct Dictionary_s));
 	memset(dict, 0, sizeof(struct Dictionary_s));
 
+	dict->line_number = 1;
+
 	/* Language and file-name stuff */
 	dict->string_set = string_set_create();
 	t = find_last_dir_separator((char *)lang);
@@ -123,8 +126,17 @@ dictionary_six_str(const char * lang,
 
 	if (NULL != affix_name)
 	{
-		/* To disable spell-checking, just set the checker to NULL */
-		dict->spell_checker = spellcheck_create(dict->lang);
+		if (dictionary_generation_request(dict))
+		{
+			const size_t initial_allocation = 256;
+			dict->num_categories_alloced = initial_allocation;
+			dict->category = malloc(sizeof(*dict->category) * initial_allocation);
+		}
+		else
+		{
+			dict->spell_checker = spellcheck_create(dict->lang);
+		}
+
 #if defined HAVE_HUNSPELL || defined HAVE_ASPELL
 		/* FIXME: Move to spellcheck-*.c */
 		if (verbosity_level(D_USER_BASIC) && (NULL == dict->spell_checker))
@@ -147,6 +159,8 @@ dictionary_six_str(const char * lang,
 			dict->macro_tag = malloc(sizeof(*dict->macro_tag));
 			memset(dict->macro_tag, 0, sizeof(*dict->macro_tag));
 		}
+
+		dict->define.set = string_id_create();
 	}
 	else
 	{
@@ -193,7 +207,8 @@ dictionary_six_str(const char * lang,
 		dict->dialect_tag.set = NULL;
 	}
 
-	dictionary_setup_locale(dict);
+	if (!dictionary_setup_defines(dict))
+		goto failure;
 
 	dict->affix_table = dictionary_six(lang, affix_name, NULL, NULL, NULL, NULL);
 	if (dict->affix_table == NULL)
@@ -231,7 +246,6 @@ dictionary_six_str(const char * lang,
 	dict->base_knowledge  = pp_knowledge_open(pp_name);
 	dict->hpsg_knowledge  = pp_knowledge_open(cons_name);
 
-	dictionary_setup_defines(dict);
 	condesc_setup(dict);
 
 	// Special-case hack.
@@ -249,6 +263,7 @@ failure:
 /**
  * Use filenames of six different files to put together the dictionary.
  */
+NO_SAN_DICT
 Dictionary
 dictionary_six(const char * lang, const char * dict_name,
                const char * pp_name, const char * cons_name,
@@ -314,10 +329,6 @@ Dictionary dictionary_create_from_file(const char * lang)
 				dictionary->dialect = NULL;
 			}
 			free(dialect_name);
-
-			if (0)
-			{
-			}
 		}
 	}
 	else
