@@ -14,14 +14,16 @@
 #include <string.h>
 
 #include "api-structures.h"             // Sentence_s (add_empty_word)
+#include "connectors.h"
+#include "dict-common/dialect.h"
 #include "dict-common/dict-affix.h"     // is_stem
 #include "dict-common/dict-common.h"
-#include "dict-common/dict-defines.h"   // SUBSCRIPT_MARK
+#include "dict-common/dict-utils.h"     // patch_subscript
 #include "dict-common/file-utils.h"
 #include "dict-common/idiom.h"
 #include "error.h"
-#include "print/print.h"
 #include "externs.h"
+#include "print/print.h"
 #include "read-dict.h"
 #include "string-set.h"
 #include "tokenize/tok-structures.h"    // MT_WALL
@@ -82,8 +84,11 @@
 
   A number following a square bracket over-rides the cost of that bracket.
   Thus, [...].5 has a cost of 0.5 while [...]2.0 has a cost of 2; that
-  is it is the same as [[...]].  Any floating point number (including
-  exponents!) is allowed.
+  is it is the same as [[...]].  Any floating point number is allowed.
+
+  Instead of a numerical cost, a symbolic cost can be used, a.k.a. a
+  "dialect component name".  The file "4.0.dialect" defines dialect names
+  (in square brackets) and their component names and values.
 
   The expression "(A+ or ())" means that you can choose either "A+" or
   the empty expression "()", that is, that the connector "A+" is
@@ -98,8 +103,8 @@
   used.
 
   The dictionary writer is not allowed to use connectors that begin in
-  "ID".  This is reserved for the connectors automatically
-  generated for idioms.
+  "_".  This is reserved for connectors that are automatically
+  generated (currently only for idioms).
 
   Dictionary words may be followed by a dot (period, "."), and a "subscript"
   identifying the word type. The subscript may be one or more letters or
@@ -107,6 +112,9 @@
   (mostly?) subscripts consisting of a single letter, and these serve mostly
   to identify the part-of-speech. In general, subscripts can also be used
   to distinguish different word senses.
+
+  Subscripts that start with "_" are reserved for words that are
+  automatically generated (currently only for idioms).
 */
 
 static bool link_advance(Dictionary dict);
@@ -132,7 +140,7 @@ static void dict_error2(Dictionary dict, const char * s, const char *s2)
 	int save_line_number    = dict->line_number;
 
 	tokens[0] = '\0';
-	for (i=0; i<5 && dict->token[0] != '\0' ; i++)
+	for (i=0; i<5 && dict->token[0] != '\0'; i++)
 	{
 		pos += snprintf(t, ERRBUFLEN, "\"%s\" ", dict->token);
 		strncat(tokens, t, ERRBUFLEN-1-pos);
@@ -150,7 +158,7 @@ static void dict_error2(Dictionary dict, const char * s, const char *s2)
 	if (s2)
 	{
 		prt_error("Error: While parsing dictionary \"%s\":\n"
-		          "%s %s\n\t Line %d, next tokens: %s\n",
+		          "%s \"%s\"\n\t Line %d, next tokens: %s\n",
 		          dict->name, s, s2, dict->line_number, tokens);
 	}
 	else
@@ -192,16 +200,6 @@ static bool get_character(Dictionary dict, int quote_mode, utf8char uc)
 		/* Skip over all comments */
 		if ((c == '%') && (!quote_mode))
 		{
-			if (0 == strncmp(dict->pin, SUPPRESS, sizeof(SUPPRESS)-1))
-			{
-				const char *nl = strchr(dict->pin + sizeof(SUPPRESS)-1, '\n');
-				if (NULL != nl)
-				{
-					dict->suppress_warning =
-						strndup(dict->pin + sizeof(SUPPRESS)-1,
-					           nl - dict->pin - sizeof(SUPPRESS) + 1);
-				}
-			}
 			while ((c != 0x0) && (c != '\n')) c = *(dict->pin++);
 			if (c == 0x0) break;
 			dict->line_number++;
@@ -235,7 +233,7 @@ static bool get_character(Dictionary dict, int quote_mode, utf8char uc)
 			uc[i] = c;
 			i++;
 		}
-		dict_error(dict, "UTF8 char is too long");
+		dict_error(dict, "UTF8 char is too long.");
 		return false;
 	}
 	uc[0] = 0x0;
@@ -306,7 +304,7 @@ static bool link_advance(Dictionary dict)
 	for (;;)
 	{
 		if (i > MAX_TOKEN_LENGTH-3) {
-			dict_error(dict, "Token too long");
+			dict_error(dict, "Token too long.");
 			return false;
 		}
 		if (quote_mode) {
@@ -319,12 +317,12 @@ static bool link_advance(Dictionary dict)
 				return true;
 			}
 			if (lg_isspace(c[0])) {
-				dict_error(dict, "White space inside of token");
+				dict_error(dict, "White space inside of token.");
 				return false;
 			}
 			if (c[0] == '\0')
 			{
-				dict_error(dict, "EOF while reading quoted token");
+				dict_error(dict, "EOF while reading quoted token.");
 				return false;
 			}
 
@@ -396,35 +394,28 @@ static bool check_connector(Dictionary dict, const char * s)
 	}
 	if (*s == '@') s++;
 	if (('h' == *s) || ('d' == *s)) s++;
-	if (!isupper((unsigned char)*s)) {
-		dict_error(dict, "Connectors must start with uppercase after an optional h or d.");
+	if (!is_connector_name_char(*s)) {
+		dict_error2(dict, "Invalid character in connector "
+		            "(connectors must start with an uppercase letter "
+		            "after an optional \"h\" or \"d\"):", (char[]){*s, '\0'});
 		return false;
 	}
-	/* Note that IDx when x is a subscript is allowed (to allow e.g. ID4id+). */
-	if ((*s == 'I') && (*(s+1) == 'D') && isupper((unsigned char)*(s+2))) {
-		dict_error(dict, "Connectors beginning with \"ID\" are forbidden");
+	if (*s == '_')
+	{
+		dict_error(dict, "Invalid character in connector "
+		           "(an initial \"_\" is reserved for internal use).");
 		return false;
 	}
 
-	bool connector_base = true;
-	s++; /* The first uppercase has been validated above. */
-	while (*(s+1)) {
-		if ((!isalnum((unsigned char)*s)) && (*s != WILD_TYPE)) {
-			dict_error(dict, "All letters of a connector must be ASCII alpha-numeric.");
+	/* The first uppercase has been validated above. */
+	do { s++; } while (is_connector_name_char(*s));
+	while (s[1]) {
+		if (!is_connector_subscript_char(*s) && (*s != WILD_TYPE)) {
+			dict_error2(dict, "Invalid character in connector subscript "
+			            "(only lowercase letters, digits, and \"*\" are allowed):",
+			            (char[]){*s, '\0'});
 			return false;
 		}
-		if (isupper((unsigned char)*s))
-		{
-			if (!connector_base)
-			{
-				dict_error(dict, "Connector subscript contains uppercase.");
-				return false;
-			}
-		}
-		else
-		{
-			connector_base = false;
-		}
 		s++;
 	}
 	return true;
@@ -489,8 +480,8 @@ NO_SAN_DICT
 static inline int dict_order_strict(const char *s, const Dict_node * dn)
 {
 	const char * t = dn->string;
-	while (*s != '\0' && *s == *t) {s++; t++;}
-	return ((*s == SUBSCRIPT_MARK)?(1):(*s))  -  ((*t == SUBSCRIPT_MARK)?(1):(*t));
+	while ((*s == *t) && (*s != '\0')) { s++; t++; }
+	return (*s - *t);
 }
 
 /**
@@ -512,7 +503,7 @@ NO_SAN_DICT
 static inline int dict_order_bare(const char *s, const Dict_node * dn)
 {
 	const char * t = dn->string;
-	while (*s != '\0' && *s == *t) {s++; t++;}
+	while ((*s == *t) && (*s != '\0')) { s++; t++; }
 	return (*s)  -  ((*t == SUBSCRIPT_MARK)?(0):(*t));
 }
 
@@ -544,7 +535,7 @@ static inline int dict_order_wild(const char * s, const Dict_node * dn)
 	const char * t = dn->string;
 
 	lgdebug(+D_DOW, "search-word='%s' dict-word='%s'\n", s, t);
-	while((*s != '\0') && (*s != SUBSCRIPT_MARK) && (*s == *t)) {s++; t++;}
+	while((*s == *t) && (*s != SUBSCRIPT_MARK) && (*s != '\0')) { s++; t++; }
 
 	if (*s == WILD_TYPE) return 0;
 
@@ -554,6 +545,13 @@ static inline int dict_order_wild(const char * s, const Dict_node * dn)
 }
 #undef D_DOW
 
+static inline Dict_node * dict_node_new(void)
+{
+	return (Dict_node*) malloc(sizeof(Dict_node));
+}
+
+/* ======================================================================== */
+#if 0
 /**
  * dict_match --  return true if strings match, else false.
  * A "bare" string (one without a subscript) will match any corresponding
@@ -564,7 +562,7 @@ static inline int dict_order_wild(const char * s, const Dict_node * dn)
  */
 static bool dict_match(const char * s, const char * t)
 {
-	while ((*s != '\0') && (*s == *t)) { s++; t++; }
+	while ((*s == *t) && (*s != '\0')) { s++; t++; }
 
 	if (*s == *t) return true; /* both are '\0' */
 	if ((*s == 0) && (*t == SUBSCRIPT_MARK)) return true;
@@ -573,13 +571,6 @@ static bool dict_match(const char * s, const char * t)
 	return false;
 }
 
-/* ======================================================================== */
-
-static inline Dict_node * dict_node_new(void)
-{
-	return (Dict_node*) malloc(sizeof(Dict_node));
-}
-
 /**
  * prune_lookup_list -- discard all list entries that don't match string
  * Walk the lookup list (of right links), discarding all nodes that do
@@ -616,15 +607,19 @@ static Dict_node * prune_lookup_list(Dict_node * restrict llist, const char * re
 	}
 	return llist;
 }
+#endif
 
 /* ======================================================================== */
 static bool subscr_match(const char *s, const Dict_node * dn)
 {
 	const char * s_sub = strrchr(s, SUBSCRIPT_MARK);
-	const char * t_sub;
+	const char * t_sub = strrchr(dn->string, SUBSCRIPT_MARK);
 
-	if (NULL == s_sub) return true;
-	t_sub = strrchr(dn->string, SUBSCRIPT_MARK);
+	if (NULL == s_sub)
+	{
+		if (NULL == t_sub) return true;
+		return !is_idiom_word(t_sub);
+	}
 	if (NULL == t_sub) return false;
 	if (0 == strcmp(s_sub, t_sub)) return true;
 
@@ -639,12 +634,11 @@ static bool subscr_match(const char *s, const Dict_node * dn)
  */
 static Dict_node *
 rdictionary_lookup(Dict_node * restrict llist,
-                   const Dict_node * restrict dn,
+                   Dict_node * restrict dn,
                    const char * restrict s,
-                   bool match_idiom,
+                   bool boolean_lookup,
                    int (*dict_order)(const char *, const Dict_node *))
 {
-	/* see comment in dictionary_lookup below */
 	int m;
 	Dict_node * dn_new;
 	if (dn == NULL) return llist;
@@ -653,19 +647,20 @@ rdictionary_lookup(Dict_node * restrict llist,
 
 	if (m >= 0)
 	{
-		llist = rdictionary_lookup(llist, dn->right, s, match_idiom, dict_order);
+		llist = rdictionary_lookup(llist, dn->right, s, boolean_lookup, dict_order);
 	}
-	if ((m == 0) && (match_idiom || !is_idiom_word(dn->string)) &&
-		 (dict_order != dict_order_wild || subscr_match(s, dn)))
+	if ((m == 0) && (dict_order != dict_order_wild || subscr_match(s, dn)))
 	{
+		if (boolean_lookup) return dn;
 		dn_new = dict_node_new();
 		*dn_new = *dn;
 		dn_new->right = llist;
+		dn_new->left = dn; /* Currently only used for inserting idioms */
 		llist = dn_new;
 	}
 	if (m <= 0)
 	{
-		llist = rdictionary_lookup(llist, dn->left, s, match_idiom, dict_order);
+		llist = rdictionary_lookup(llist, dn->left, s, boolean_lookup, dict_order);
 	}
 	return llist;
 }
@@ -674,8 +669,6 @@ rdictionary_lookup(Dict_node * restrict llist,
  * file_lookup_list() - return list of words in the file-backed dictionary.
  *
  * Returns a pointer to a lookup list of the words in the dictionary.
- * Matches include words that appear in idioms.  To exclude idioms, use
- * abridged_lookup_list() to obtain matches.
  *
  * This list is made up of Dict_nodes, linked by their right pointers.
  * The node, file and string fields are copied from the dictionary.
@@ -684,18 +677,12 @@ rdictionary_lookup(Dict_node * restrict llist,
  */
 Dict_node * file_lookup_list(const Dictionary dict, const char *s)
 {
-	Dict_node * llist =
-		rdictionary_lookup(NULL, dict->root, s, true, dict_order_bare);
-	llist = prune_lookup_list(llist, s);
-	return llist;
+	return rdictionary_lookup(NULL, dict->root, s, false, dict_order_bare);
 }
 
 bool file_boolean_lookup(Dictionary dict, const char *s)
 {
-	Dict_node *llist = file_lookup_list(dict, s);
-	bool boool = (llist != NULL);
-	file_free_lookup(llist);
-	return boool;
+	return !!rdictionary_lookup(NULL, dict->root, s, true, dict_order_bare);
 }
 
 void file_free_lookup(Dict_node *llist)
@@ -726,25 +713,23 @@ void free_insert_list(Dict_node *ilist)
  */
 Dict_node * file_lookup_wild(Dictionary dict, const char *s)
 {
-	bool lookup_idioms = test_enabled("lookup-idioms");
 	char * ds = strrchr(s, SUBSCRIPT_DOT); /* Only the rightmost dot is a
 	                                          candidate for SUBSCRIPT_DOT */
 	char * ws = strrchr(s, WILD_TYPE);     /* A SUBSCRIPT_DOT can only appear
                                              after a wild-card */
 	Dict_node * result;
-	char * stmp = strdup(s);
+	char * stmp = strdupa(s);
 
 	/* It is not a SUBSCRIPT_DOT if it is at the end or before the wild-card.
 	 * E.g: "Dr.", "i.*", "." */
 	if ((NULL != ds) && ('\0' != ds[1]) && ((NULL == ws) || (ds > ws)))
 		stmp[ds-s] = SUBSCRIPT_MARK;
 
-	result =
-	 rdictionary_lookup(NULL, dict->root, stmp, lookup_idioms, dict_order_wild);
-	free(stmp);
+	result = rdictionary_lookup(NULL, dict->root, stmp, false, dict_order_wild);
 	return result;
 }
 
+#if 0
 /**
  * abridged_lookup_list() - return lookup list of words in the dictionary
  *
@@ -759,17 +744,14 @@ Dict_node * file_lookup_wild(Dictionary dict, const char *s)
  */
 static Dict_node * abridged_lookup_list(const Dictionary dict, const char *s)
 {
-	Dict_node *llist;
-	llist = rdictionary_lookup(NULL, dict->root, s, false, dict_order_bare);
-	llist = prune_lookup_list(llist, s);
-	return llist;
+	return rdictionary_lookup(NULL, dict->root, s, false, dict_order_bare);
 }
+#endif
 
 /**
  * strict_lookup_list() - return exact match in the dictionary
  *
  * Returns a pointer to a lookup list of the words in the dictionary.
- * Excludes any idioms that contain the word.
  *
  * This list is made up of Dict_nodes, linked by their right pointers.
  * The node, file and string fields are copied from the dictionary.
@@ -781,30 +763,28 @@ static Dict_node * abridged_lookup_list(const Dictionary dict, const char *s)
  */
 static Dict_node * strict_lookup_list(const Dictionary dict, const char *s)
 {
-	Dict_node *llist;
-	llist = rdictionary_lookup(NULL, dict->root, s, false, dict_order_strict);
-	llist = prune_lookup_list(llist, s);
-	return llist;
+	return rdictionary_lookup(NULL, dict->root, s, false, dict_order_strict);
 }
 
 /* ======================================================================== */
 /**
  * Allocate a new Exp node.
  */
-Exp *Exp_create(Dictionary dict)
+Exp *Exp_create(Pool_desc *mp)
 {
-	Exp *e = pool_alloc(dict->Exp_pool);
+	Exp *e = pool_alloc(mp);
+	e->tag_type = Exptag_none;
 	return e;
 }
 
 /**
  * Duplicate the given Exp node.
- * This is needed in case it is to be participate more than once in a
+ * This is needed in case it participates more than once in a
  * single expression.
  */
-static Exp *Exp_create_dup(Dictionary dict, Exp *old_e)
+Exp *Exp_create_dup(Pool_desc *mp, Exp *old_e)
 {
-	Exp *new_e = Exp_create(dict);
+	Exp *new_e = Exp_create(mp);
 
 	*new_e = *old_e;
 
@@ -815,9 +795,9 @@ static Exp *Exp_create_dup(Dictionary dict, Exp *old_e)
  * This creates a node with zero children.  Initializes
  * the cost to zero.
  */
-static Exp * make_zeroary_node(Dictionary dict)
+static Exp * make_zeroary_node(Pool_desc *mp)
 {
-	Exp * n = Exp_create(dict);
+	Exp * n = Exp_create(mp);
 	n->type = AND_type;  /* these must be AND types */
 	n->cost = 0.0;
 	n->operand_first = NULL;
@@ -829,10 +809,10 @@ static Exp * make_zeroary_node(Dictionary dict)
  * This creates a node with one child (namely e).  Initializes
  * the cost to zero.
  */
-static Exp * make_unary_node(Dictionary dict, Exp * e)
+Exp *make_unary_node(Pool_desc *mp, Exp * e)
 {
 	Exp * n;
-	n = Exp_create(dict);
+	n = Exp_create(mp);
 	n->type = AND_type;  /* these must be AND types */
 	n->operand_next = NULL;
 	n->cost = 0.0;
@@ -844,11 +824,11 @@ static Exp * make_unary_node(Dictionary dict, Exp * e)
  * Create an AND_type expression. The expressions nl, nr will be
  * AND-ed together.
  */
-static Exp * make_and_node(Dictionary dict, Exp* nl, Exp* nr)
+static Exp * make_and_node(Pool_desc *mp, Exp* nl, Exp* nr)
 {
 	Exp* n;
 
-	n = Exp_create(dict);
+	n = Exp_create(mp);
 	n->type = AND_type;
 	n->operand_next = NULL;
 	n->cost = 0.0;
@@ -860,14 +840,14 @@ static Exp * make_and_node(Dictionary dict, Exp* nl, Exp* nr)
 	return n;
 }
 
-static Exp *make_op_Exp(Dictionary dict, Exp_type t)
+static Exp *make_op_Exp(Pool_desc *mp, Exp_type t)
 {
-	Exp * n = Exp_create(dict);
+	Exp * n = Exp_create(mp);
 	n->type = t;
 	n->operand_next = NULL;
 	n->cost = 0.0;
 
-	/* The caller is supposed to assign n->operand. */
+	/* The caller is supposed to assign n->operand->first. */
 	return n;
 }
 
@@ -875,11 +855,11 @@ static Exp *make_op_Exp(Dictionary dict, Exp_type t)
  * Create an OR_type expression. The expressions nl, nr will be
  * OR-ed together.
  */
-static Exp * make_or_node(Dictionary dict, Exp* nl, Exp* nr)
+static Exp * make_or_node(Pool_desc *mp, Exp* nl, Exp* nr)
 {
 	Exp* n;
 
-	n = Exp_create(dict);
+	n = Exp_create(mp);
 	n->type = OR_type;
 	n->operand_next = NULL;
 	n->cost = 0.0;
@@ -896,9 +876,9 @@ static Exp * make_or_node(Dictionary dict, Exp* nl, Exp* nr)
  * and the other as zeroary node.  This has the effect of creating
  * what used to be called an optional node.
  */
-static Exp *make_optional_node(Dictionary dict, Exp *e)
+static Exp *make_optional_node(Pool_desc *mp, Exp *e)
 {
-	return make_or_node(dict, make_zeroary_node(dict), e);
+	return make_or_node(mp, make_zeroary_node(mp), e);
 }
 
 /**
@@ -909,7 +889,7 @@ static Exp *make_optional_node(Dictionary dict, Exp *e)
  */
 static Exp * make_dir_connector(Dictionary dict, int i)
 {
-	Exp* n = Exp_create(dict);
+	Exp* n = Exp_create(dict->Exp_pool);
 	char *constring;
 
 	n->dir = dict->token[i];
@@ -935,6 +915,29 @@ static Exp * make_dir_connector(Dictionary dict, int i)
 }
 
 /* ======================================================================== */
+/**
+ * Add an optional macro/word tag, for expression debugging.
+ * Enabled by !test="macro-tag". This tag is used only in expression printing.
+ */
+static unsigned int exptag_macro_add(Dictionary dict, const char *tag)
+{
+	expression_tag *mt = dict->macro_tag;
+	if (mt == NULL) return 0;
+
+	if (mt->num == mt->size)
+	{
+		if (mt->num == 0)
+			mt->size = 128;
+		else
+			mt->size *= 2;
+
+		mt->name = realloc(mt->name, mt->size * sizeof(*mt->name));
+	}
+	mt->name[mt->num] = tag;
+
+	return mt->num++;
+}
+
 /**
  * make_connector() -- make a node for a connector or dictionary word.
  *
@@ -943,7 +946,7 @@ static Exp * make_dir_connector(Dictionary dict, int i)
 static Exp * make_connector(Dictionary dict)
 {
 	Exp * n;
-	Dict_node *dn, *dn_head;
+	Dict_node *dn;
 	int i;
 
 	i = strlen(dict->token) - 1;  /* this must be +, - or $ if a connector */
@@ -953,25 +956,30 @@ static Exp * make_connector(Dictionary dict)
 	{
 		/* If we are here, token is a word */
 		patch_subscript(dict->token);
-		dn_head = abridged_lookup_list(dict, dict->token);
-		dn = dn_head;
-		while ((dn != NULL) && (strcmp(dn->string, dict->token) != 0))
-		{
-			dn = dn->right;
-		}
+		dn = strict_lookup_list(dict, dict->token);
 		if (dn == NULL)
 		{
-			file_free_lookup(dn_head);
-			dict_error(dict, "Perhaps missing + or - in a connector.\n"
+			file_free_lookup(dn);
+			dict_error2(dict, "Perhaps missing + or - in a connector.\n"
 			                 "Or perhaps you forgot the subscript on a word.\n"
-			                 "Or perhaps a word is used before it is defined.");
+			                 "Or perhaps the word is used before it is defined:",
+			                 dict->token);
+			return NULL;
+		}
+		if (dn->right != NULL)
+		{
+			file_free_lookup(dn);
+			dict_error2(dict, "Referencing a duplicate word:", dict->token);
+			/* Note: A word which becomes duplicate latter evades this check. */
 			return NULL;
 		}
 
-		/* Wrap it in a unary node as a placeholder for a cost if needed. */
-		n = make_unary_node(dict, dn->exp);
+		/* Wrap it in a unary node as a placeholder for a macro tag and cost. */
+		n = make_unary_node(dict->Exp_pool, dn->exp);
+		n->tag_id = exptag_macro_add(dict, dn->string);
+		if (n->tag_id != 0) n->tag_type = Exptag_macro;
 
-		file_free_lookup(dn_head);
+		file_free_lookup(dn);
 	}
 	else
 	{
@@ -998,7 +1006,7 @@ static Exp * make_connector(Dictionary dict)
 			min = make_dir_connector(dict, i);
 			if (NULL == min) return NULL;
 
-			n = make_or_node(dict, plu, min);
+			n = make_or_node(dict->Exp_pool, plu, min);
 		}
 		else
 		{
@@ -1018,42 +1026,6 @@ static Exp * make_connector(Dictionary dict)
 /* ======================================================================== */
 /* Empty-word handling. */
 
-/** Expression-creating function versions that use the Sentence
- *  expression memory pools.
- *  (FIXME: Unify with the similar functions above.)
- */
-static Exp *make_or_node_from_pool(Sentence sent, Exp *nl, Exp *nr)
-{
-	Exp* n;
-
-	n = pool_alloc(sent->Exp_pool);
-	n->type = OR_type;
-	n->operand_next = NULL;
-	n->cost = 0.0;
-
-	n->operand_first = nl;
-	nl->operand_next = nr;
-	nr->operand_next = NULL;
-
-	return n;
-}
-
-static Exp *make_zeroary_node_from_pool(Sentence sent)
-{
-	Exp * n = pool_alloc(sent->Exp_pool);
-	n->type = AND_type;  /* these must be AND types */
-	n->operand_next = NULL;
-	n->cost = 0.0;
-	n->operand_first = NULL;
-
-	return n;
-}
-
-static Exp *make_optional_node_from_pool(Sentence sent, Exp *e)
-{
-	return make_or_node_from_pool(sent, make_zeroary_node_from_pool(sent), e);
-}
-
 /** Insert ZZZ+ connectors.
  *  This function was mainly used to support using empty-words, a concept
  *  that has been eliminated. However, it is still used to support linking of
@@ -1077,17 +1049,17 @@ void add_empty_word(Sentence sent, X_node *x)
 		//lgdebug(+0, "Processing '%s'\n", x->string);
 
 		/* zn points at {ZZZ+} */
-		zn = pool_alloc(sent->Exp_pool);
+		zn = Exp_create(sent->Exp_pool);
 		zn->dir = '+';
 		zn->condesc = condesc_add(&sent->dict->contable, ZZZ);
 		zn->multi = false;
 		zn->type = CONNECTOR_type;
 		zn->operand_next = NULL; /* unused, but to be on the safe side */
 		zn->cost = 0.0;
-		zn = make_optional_node_from_pool(sent, zn);
+		zn = make_optional_node(sent->Exp_pool, zn);
 
 		/* an will be {ZZZ+} & (plain-word-exp) */
-		an = pool_alloc(sent->Exp_pool);
+		an = Exp_create(sent->Exp_pool);
 		an->type = AND_type;
 		an->operand_next = NULL;
 		an->cost = 0.0;
@@ -1106,6 +1078,7 @@ void add_empty_word(Sentence sent, X_node *x)
  */
 static bool is_number(const char * str)
 {
+	if (str[0] == '\0') return false; /* End of file. */
 	if ('+' == str[0] || '-' == str[0]) str++;
 	size_t numlen = strspn(str, "0123456789.");
 
@@ -1161,7 +1134,7 @@ static Exp *make_expression(Dictionary dict)
 			if (!link_advance(dict)) {
 				return NULL;
 			}
-			nl = make_optional_node(dict, nl);
+			nl = make_optional_node(dict->Exp_pool, nl);
 		}
 		else if (is_equal(dict, '['))
 		{
@@ -1180,14 +1153,15 @@ static Exp *make_expression(Dictionary dict)
 				return NULL;
 			}
 
-			/* A square bracket can have a number after it.
-			 * If it is present, then that number is interpreted
-			 * as the cost of the bracket. Else, the cost of a
+			/* A square bracket can have a number or a name after it.
+			 * If a number is present, then that it is interpreted
+			 * as the cost of the bracket. If a name is present, it
+			 * is used as an expression tag. Else, the cost of a
 			 * square bracket is 1.0.
 			 */
 			if (is_number(dict->token))
 			{
-				double cost;
+				float cost;
 
 				if (strtodC(dict->token, &cost))
 				{
@@ -1202,6 +1176,28 @@ static Exp *make_expression(Dictionary dict)
 					return NULL;
 				}
 			}
+			else if ((strcmp(dict->token, "or") != 0) &&
+			         (strcmp(dict->token, "and") != 0) &&
+			         isalpha(dict->token[0]))
+			{
+				const char *bad = valid_dialect_name(dict->token);
+				if (bad != NULL)
+				{
+					char badchar[] = { *bad, '\0' };
+					dict_error2(dict, "Invalid character in dialect tag name:",
+					           badchar);
+					return NULL;
+				}
+				if (nl->tag_type != Exptag_none)
+				{
+					nl = make_unary_node(dict->Exp_pool, nl);
+				}
+				nl->tag_id = exptag_dialect_add(dict, dict->token);
+				nl->tag_type = Exptag_dialect;
+				if (!link_advance(dict)) {
+					return NULL;
+				}
+			}
 			else
 			{
 				nl->cost += 1.0;
@@ -1217,7 +1213,7 @@ static Exp *make_expression(Dictionary dict)
 		else if (is_equal(dict, ')') || is_equal(dict, ']'))
 		{
 			/* allows "()" or "[]" */
-			nl = make_zeroary_node(dict);
+			nl = make_zeroary_node(dict->Exp_pool);
 		}
 		else
 		{
@@ -1230,13 +1226,13 @@ static Exp *make_expression(Dictionary dict)
 			/* Part 2/2 of SYM_AND processing. */
 
 			/* Expand A ^ B into the expr ((A & B) or (B & A)). */
-			Exp *na = make_and_node(dict,
-			                   Exp_create_dup(dict, e_tail),
-			                   Exp_create_dup(dict, nl));
-			Exp *nb = make_and_node(dict,
-			                   Exp_create_dup(dict, nl),
-			                   Exp_create_dup(dict, e_tail));
-			Exp *or = make_or_node(dict, na, nb);
+			Exp *na = make_and_node(dict->Exp_pool,
+			                   Exp_create_dup(dict->Exp_pool, e_tail),
+			                   Exp_create_dup(dict->Exp_pool, nl));
+			Exp *nb = make_and_node(dict->Exp_pool,
+			                   Exp_create_dup(dict->Exp_pool, nl),
+			                   Exp_create_dup(dict->Exp_pool, e_tail));
+			Exp *or = make_or_node(dict->Exp_pool, na, nb);
 
 			*e_tail = *or; /* SYM_AND result */
 			is_sym_and = false;
@@ -1281,14 +1277,14 @@ static Exp *make_expression(Dictionary dict)
 		 * expression level. */
 		if (e_head == NULL)
 		{
-			e_head = make_op_Exp(dict, op);
+			e_head = make_op_Exp(dict->Exp_pool, op);
 			e_head->operand_first = nl;
 		}
 		else
 		{
 			if (e_head->type != op)
 			{
-				dict_error(dict, "\"and\" and \"or\" at the same level in an expression");
+				dict_error(dict, "\"and\" and \"or\" at the same level in an expression.");
 				return NULL;
 			}
 		}
@@ -1391,7 +1387,7 @@ static Dict_node * dsw_vine_to_tree (Dict_node *root, int size)
 	vine_head.right = root;
 
 	dsw_compression(&vine_head, size - full_count);
-	for (size = full_count ; size > 1 ; size /= 2)
+	for (size = full_count; size > 1; size /= 2)
 	{
 		dsw_compression(&vine_head, size / 2);
 	}
@@ -1399,6 +1395,53 @@ static Dict_node * dsw_vine_to_tree (Dict_node *root, int size)
 }
 
 /* ======================================================================== */
+/**
+ * Notify about a duplicate word, unless allowed or it is an idiom definition.
+ * Idioms are exempt because historically they couldn't be
+ * differentiated using a subscript if duplicate definitions were
+ * convenience (and also they were not inserted into the dictionary so
+ * their duplicate check got neglected).
+ *
+ * The following dictionary definition allows duplicate words:
+ * #define allow-duplicate-words true
+ * An idiom duplicate check can be requested using the test "dup-idioms".
+ */
+static bool dup_word_error(Dictionary dict, Dict_node *newnode)
+{
+	static int dup_idioms = -1;
+	static int allow_duplicate_words = -1;
+
+	if (allow_duplicate_words == 1) return false;
+	if (allow_duplicate_words == -1)
+	{
+		const char *s = linkgrammar_get_dict_define(dict, "allow-duplicate-words");
+
+		allow_duplicate_words = ((s != NULL) && (0 == strcasecmp(s, "true")));
+		if (allow_duplicate_words) return false;
+		if (dup_idioms == -1) dup_idioms = !!test_enabled("dup-idioms");
+	}
+
+	/* FIXME: Make central. */
+	static Exp null_exp =
+	{
+		.type = AND_type,
+		.operand_first = NULL,
+		.operand_next = NULL,
+	};
+
+	if (!contains_underbar(newnode->string) || dup_idioms)
+	{
+		dict_error2(dict, "Ignoring word which has been multiply defined:",
+		            newnode->string);
+		/* Too late to skip insertion - insert it with a null expression. */
+		newnode->exp = &null_exp;
+
+		return true;
+	}
+
+	return false;
+}
+
 /**
  * Insert the new node into the dictionary below node n.
  * "newnode" left and right fields are NULL, and its string is already
@@ -1413,27 +1456,10 @@ Dict_node *insert_dict(Dictionary dict, Dict_node *n, Dict_node *newnode)
 {
 	if (NULL == n) return newnode;
 
-	static Exp null_exp =
-	{
-		.type = AND_type,
-		.operand_first = NULL,
-		.operand_next = NULL,
-	};
 	int comp = dict_order_strict(newnode->string, n);
 
-	if (0 == comp &&
-	    /* Suppress reporting duplicate idioms until they are fixed. */
-	    (!contains_underbar(newnode->string) || test_enabled("dup-idioms")))
-	{
-		char t[80+MAX_TOKEN_LENGTH];
-		snprintf(t, sizeof(t),
-		         "Ignoring word \"%s\", which has been multiply defined:",
-		         newnode->string);
-		dict_error(dict, t);
-		/* Too late to skip insertion - insert it with a null expression. */
-		newnode->exp = &null_exp;
-		comp = -1;
-	}
+	if ((0 == comp) && dup_word_error(dict, newnode))
+	    comp = -1;
 
 	if (comp < 0)
 	{
@@ -1458,17 +1484,6 @@ Dict_node *insert_dict(Dictionary dict, Dict_node *n, Dict_node *newnode)
 	/* return rebalance(n); Uncomment to get an AVL tree */
 }
 
-/**
- * Find if a warning symbol exists in the currently suppress list.
- * The warning symbols are constructed in a way that disallow overlap
- * matching.
- */
-static bool is_warning_suppressed(Dictionary dict, const char *warning_symbol)
-{
-	if (NULL == dict->suppress_warning) return false;
-	return (NULL != strstr(dict->suppress_warning, warning_symbol));
-}
-
 /**
  * Remember the length_limit definitions in a list according to their order.
  * The order is kept to allow later more specific definitions to override
@@ -1538,7 +1553,7 @@ static void insert_length_limit(Dictionary dict, Dict_node *dn)
  */
 void insert_list(Dictionary dict, Dict_node * p, int l)
 {
-	Dict_node * dn, *dn_head, *dn_second_half;
+	Dict_node * dn, *dn_second_half;
 	int k, i; /* length of first half */
 
 	if (l == 0) return;
@@ -1554,10 +1569,11 @@ void insert_list(Dictionary dict, Dict_node * p, int l)
 	dn_second_half = dn->left;
 	dn->left = dn->right = NULL;
 
-	if (is_idiom_word(dn->string))
+	const char *sm = strchr(dn->string, SUBSCRIPT_MARK);
+	if ((NULL != sm) && ('_' == sm[1]))
 	{
 		prt_error("Warning: Word \"%s\" found near line %d of \"%s\".\n"
-		        "\tWords ending \".Ix\" (x a number) are reserved for idioms.\n"
+		        "\tWords ending \"._\" are reserved for internal use.\n"
 		        "\tThis word will be ignored.\n",
 		        dn->string, dict->line_number, dict->name);
 		free(dn);
@@ -1572,50 +1588,100 @@ void insert_list(Dictionary dict, Dict_node * p, int l)
 		dict->root = insert_dict(dict, dict->root, dn);
 		insert_length_limit(dict, dn);
 		dict->num_entries++;
-
-		if ((verbosity_level(D_DICT+0) && !is_warning_suppressed(dict, DUP_BASE)) ||
-		    verbosity_level(D_SPEC+3))
-		{
-			/* Warn if there are words with a subscript that match a bare word. */
-			const char *sm = strchr(dn->string, SUBSCRIPT_MARK);
-			char *bareword;
-
-			if (NULL != sm)
-			{
-				bareword = strdupa(dn->string);
-				bareword[sm - dn->string] = '\0';
-			}
-			else
-			{
-				bareword = (char *)dn->string;
-			}
-
-			if ((dn_head = strict_lookup_list(dict, bareword)) != NULL)
-			{
-				bool match_found = false;
-				for (Dict_node *dnx = dn_head; dnx != NULL; dnx = dnx->right) {
-					if (0 != strcmp(dnx->string, dn->string))
-					{
-						if (!match_found)
-						{
-							prt_error("Warning: The word \"%s\" found near line "
-										 "%d of \"%s\"\n\t matches the following words:",
-										 dn->string, dict->line_number, dict->name);
-							match_found = true;
-						}
-						prt_error("\t%s", dnx->string);
-					}
-				}
-				if (match_found) prt_error("\n");
-				file_free_lookup(dn_head);
-			}
-		}
 	}
 
 	insert_list(dict, p, k);
 	insert_list(dict, dn_second_half, l-k-1);
 }
 
+static void add_define(Dictionary dict, const char *name, const char *value)
+{
+	int id = string_id_add(name, dict->define.set);
+
+	if (!IS_DB_DICT(dict) && (dict->define.size >= (unsigned int)id))
+	{
+		prt_error("Warning: Redefinition of \"%s\", "
+		          "found near line %d of \"%s\"\n",
+		          name, dict->line_number, dict->name);
+	}
+	else
+	{
+		dict->define.size++;
+		dict->define.value =
+			realloc(dict->define.value, dict->define.size * sizeof(char *));
+		dict->define.name =
+			realloc(dict->define.name, dict->define.size * sizeof(char *));
+		assert(dict->define.size == (unsigned int)id,
+		       "\"define\" array size inconsistency");
+		dict->define.name[id - 1] = string_set_add(name, dict->string_set);
+	}
+	dict->define.value[id - 1] = string_set_add(value, dict->string_set);
+}
+
+static bool is_directive(const char *s)
+{
+	return
+		(strcmp(s, UNLIMITED_CONNECTORS_WORD) == 0) ||
+		(strncmp(s, LIMITED_CONNECTORS_WORD, sizeof(LIMITED_CONNECTORS_WORD)-1) == 0);
+}
+
+static bool is_correction(const char *s)
+{
+	static const char correction_mark[] = { SUBSCRIPT_MARK, '#' , '\0'};
+	return strstr(s, correction_mark) != 0;
+}
+
+static void add_category(Dictionary dict, Exp *e, Dict_node *dn, int n)
+{
+	if (n == 1)
+	{
+		if (is_macro(dn->string)) return;
+		if (!dict->generate_walls && is_wall(dn->string)) return;
+		if (is_correction(dn->string)) return;
+		if (is_directive(dn->string)) return;
+	}
+
+	/* Add a category with a place for n words. */
+	dict->num_categories++;
+	if (dict->num_categories >= dict->num_categories_alloced)
+	{
+		dict->num_categories_alloced *= 2;
+		dict->category =
+			realloc(dict->category,
+			        sizeof(*dict->category) * dict->num_categories_alloced);
+	}
+	dict->category[dict->num_categories].word =
+		malloc(sizeof(dict->category[0].word) * n);
+
+	n = 0;
+	for (Dict_node *dnx = dn; dnx != NULL; dnx = dnx->left)
+	{
+		if (is_macro(dnx->string)) continue;
+		if (!dict->generate_walls && is_wall(dnx->string)) continue;
+		if (is_correction(dnx->string)) continue;
+		if (is_directive(dnx->string)) return;
+		dict->category[dict->num_categories].word[n] = dnx->string;
+		n++;
+	}
+
+	if (n == 0)
+	{
+		free(dict->category[dict->num_categories].word);
+		--dict->num_categories;
+	}
+	else
+	{
+		assert(dict->num_categories < 1024 * 1024, "Insane number of categories");
+		char category_string[16]; /* For the tokenizer - not used here */
+		snprintf(category_string, sizeof(category_string), " %x",
+		         dict->num_categories);
+		string_set_add(category_string, dict->string_set);
+		dict->category[dict->num_categories].exp = e;
+		dict->category[dict->num_categories].num_words = n;
+		dict->category[dict->num_categories].name = "";
+	}
+}
+
 /**
  * read_entry() -- read one dictionary entry
  * Starting with the current token, parse one dictionary entry.
@@ -1714,6 +1780,22 @@ static bool read_entry(Dictionary dict)
 
 			return true;
 		}
+		else if (0 == strcmp(dict->token, "#define"))
+		{
+			if (!link_advance(dict)) goto syntax_error;
+			const char *name = strdupa(dict->token);
+
+			/* Get the value. */
+			if (!link_advance(dict)) goto syntax_error;
+			add_define(dict, name, dict->token);
+
+			if (!link_advance(dict)) goto syntax_error;
+			if (!is_equal(dict, ';'))
+			{
+				dict_error(dict, "Expecting \";\" at the end of #define.");
+				goto syntax_error;
+			}
+		}
 		else
 		{
 			Dict_node * dn_new = dict_node_new();
@@ -1751,21 +1833,24 @@ static bool read_entry(Dictionary dict)
 		goto syntax_error;
 	}
 
+	if (dn == NULL)
+	{
+		dict_error(dict, "Expecting a token before \":\".");
+		goto syntax_error;
+	}
+
 	/* At this point, dn points to a list of Dict_nodes connected by
-	 * their left pointers. These are to be inserted into the dictionary */
+	 * their left pointers. These are to be inserted into the dictionary. */
 	i = 0;
 	for (dnx = dn; dnx != NULL; dnx = dnx->left)
 	{
 		dnx->exp = n;
 		i++;
 	}
-	dict->insert_entry(dict, dn, i);
+	if (IS_GENERATION(dict))
+		add_category(dict, n, dn, i);
 
-	if (dict->suppress_warning)
-	{
-		free((void *)dict->suppress_warning);
-		dict->suppress_warning = NULL;
-	}
+	dict->insert_entry(dict, dn, i);
 
 	/* pass the ; */
 	if (!link_advance(dict))
@@ -1781,12 +1866,22 @@ syntax_error:
 	return false;
 }
 
+void print_dictionary_defines(Dictionary dict)
+{
+	for (size_t i = 0; i < dict->define.size; i++)
+	{
+		const char *value = dict->define.value[i];
+		int q = (int)(strcspn(value, SPECIAL) == strlen(value));
+		printf("#define %s %s%s%s\n",
+		       dict->define.name[i], &"\""[q], value, &"\""[q]);
+	}
+}
 
 static void rprint_dictionary_data(Dict_node * n)
 {
 	if (n == NULL) return;
 	rprint_dictionary_data(n->left);
-	printf("%s: %s\n", n->string, lg_exp_stringify(n->exp));
+	printf("%s: %s\n", n->string, exp_stringify(n->exp));
 	rprint_dictionary_data(n->right);
 }
 
@@ -1817,6 +1912,18 @@ bool read_dictionary(Dictionary dict)
 			return false;
 		}
 	}
+
+	if (dict->category != NULL)
+	{
+		/* Create a category element which contains 0 words, to signify the
+		 * end of the category array for the user API. The number of
+		 * categories will not get changed because macros are considered an
+		 * invalid word. */
+		Exp dummy_exp;
+		add_category(dict, &dummy_exp, NULL, 0);
+		dict->category[dict->num_categories + 1].num_words = 0;
+	}
+
 	dict->root = dsw_tree_to_vine(dict->root);
 	dict->root = dsw_vine_to_tree(dict->root, dict->num_entries);
 	return true;
