@@ -30,9 +30,23 @@ struct Disjunct_struct
 	/* 48 bytes of common stuff. */
 	Disjunct *next;
 	Connector *left, *right;
-	double cost;
 	gword_set *originating_gword; /* Set of originating gwords */
-	const char *word_string;      /* Subscripted dictionary word */
+	union
+	{
+		struct
+		{
+			unsigned int is_category; /* Use the category field */
+			float cost;               /* Disjunct cost */
+			const char *word_string;  /* Subscripted dictionary word */
+		};
+		struct
+		{
+			/* Dictionary category & disjunct cost (for sentence generation). */
+			unsigned int num_categories;
+			unsigned int num_categories_alloced;
+			Category_cost *category;
+		};
+	};
 
 	/* Shared by different steps. For what | when. */
 	union
@@ -45,7 +59,7 @@ struct Disjunct_struct
 	union
 	{
 		uint32_t dup_hash;        /* Duplicate elimination | before pruning */
-		uint32_t unused2;         /* Unused now | before and during parsing */
+		int32_t ordinal;          /* Generation mode | after d. elimination */
 	}; /* 4 bytes */
 
 	struct
@@ -59,11 +73,12 @@ struct Disjunct_struct
 
 /* Disjunct utilities ... */
 void free_disjuncts(Disjunct *);
-void free_sentence_disjuncts(Sentence);
+void free_sentence_disjuncts(Sentence, bool);
+void free_categories(Sentence);
+void free_categories_from_disjunct_array(Disjunct *, unsigned int);
 unsigned int count_disjuncts(Disjunct *);
 Disjunct * catenate_disjuncts(Disjunct *, Disjunct *);
-Disjunct * eliminate_duplicate_disjuncts(Disjunct *);
-char * print_one_disjunct(Disjunct *);
+Disjunct * eliminate_duplicate_disjuncts(Disjunct *, bool);
 int left_connector_count(Disjunct *);
 int right_connector_count(Disjunct *);
 
@@ -72,8 +87,13 @@ Tracon_sharing *pack_sentence_for_parsing(Sentence);
 void free_tracon_sharing(Tracon_sharing *);
 void count_disjuncts_and_connectors(Sentence, unsigned int *, unsigned int *);
 
-void print_connector_list(Connector *, const char *);
-void print_disjunct_list(Disjunct *, const char *);
+/* Print disjunct/connector */
+char *print_one_disjunct_str(const Disjunct *);
+char *print_one_connector_str(const Connector *, const char *);
+char *print_connector_list_str(const Connector *, const char *);
+void print_one_connector(const Connector *, const char *);
+void print_connector_list(const Connector *, const char *);
+void print_disjunct_list(const Disjunct *, const char *);
 void print_all_disjuncts(Sentence);
 
 /* Save and restore sentence disjuncts */
@@ -105,7 +125,6 @@ typedef struct
 
 typedef struct
 {
-	unsigned int *num_cnctrs_per_word[2]; /* Indexed by word number */
 	/* Table of tracons. A 32bit index into the connector array in
 	 * memblock (instead of (Connector*)) is used for better use of the
 	 * CPU cache on 64-bit CPUs. */
@@ -134,6 +153,14 @@ struct tracon_sharing_s
 	int word_offset;            /* Start number for connector tracon_id */
 	bool is_pruning;            /* false: Parsing step, true: Pruning step */
 	Tracon_list *tracon_list;   /* Used only for pruning */
+
+	/* The number of different uppercase connector parts per side / word,
+	 * for sizing the prune power table. */
+	uint8_t *uc_seen[2];        /* The last word number in which an
+										  * uppercase connector part has been seen,
+										  * (hence doesn't need clearing between words).
+										  * Indexed by uc_num. */
+	unsigned int *num_cnctrs_per_word[2]; /* Indexed by word number */
 };
 
 void *save_disjuncts(Sentence, Tracon_sharing *);
