@@ -15,7 +15,7 @@
 #include "dict-common/regex-morph.h"    // match_regex
 #include "connectors.h"                 // MAX_SENTENCE
 #include "disjunct-utils.h"             // Disjunct_struct
-#include "lg_assert.h"
+#include "error.h"
 #include "linkage.h"
 #include "sane.h"
 #include "tokenize/tok-structures.h"    // Wordgraph_pathpos_s
@@ -94,8 +94,8 @@ static void wordgraph_path_append(Wordgraph_pathpos **nwp, const Gword **path,
 	}
 	else
 	{
-		lgdebug(D_WPA, "Path position to be replaced (len %zu): %zu\n", n,
-		                wpt - *nwp);
+		lgdebug(D_WPA, "Path position to be replaced (len %zu): %d\n", n,
+		                (int)(wpt - *nwp));
 		n = wpt - *nwp; /* Replace this path. */
 	}
 	(*nwp)[n].word = p;
@@ -209,8 +209,11 @@ static size_t num_islands(const Linkage lkg, const Gword **wg_path)
 		/* Skip null words which are optional words. */
 		if ((NULL == *wg_path) || ((*wg_path)->sent_wordidx != w))
 		{
-			assert(word[w].prev == word[w].next);
-			assert((NULL == cdj[w]) && lkg->sent->word[w].optional);
+			assert(word[w].prev == word[w].next,
+			   "A skipped optional word found in an island");
+			assert((NULL == cdj[w]) && lkg->sent->word[w].optional,
+			   "A matching disjunct found for a skipped optional word");
+
 
 			word[w].prev = NO_WORD;
 			word[w].inum = -1; /* not belonging to any island */
@@ -285,13 +288,13 @@ bool sane_linkage_morphism(Sentence sent, Linkage lkg, Parse_Options opts)
 {
 	Wordgraph_pathpos *wp_new = NULL;
 	Wordgraph_pathpos *wp_old = NULL;
-	Wordgraph_pathpos *wpp;
+	Wordgraph_pathpos *wpp = NULL;
 	Gword **next; /* next Wordgraph words of the current word */
 	size_t i;
-	size_t null_count_found = 0;
+	unsigned int null_count_found = 0;
 
 	bool match_found = true; /* if all the words are null - it's still a match */
-	Gword **lwg_path;
+	Gword **lwg_path = NULL;
 
 	Dictionary afdict = sent->dict->affix_table;       /* for SANEMORPHISM */
 	char *const affix_types = alloca(sent->length*2 + 1);   /* affix types */
@@ -372,7 +375,7 @@ bool sane_linkage_morphism(Sentence sent, Linkage lkg, Parse_Options opts)
 				if ((null_count_found > lkg->sent->null_count) &&
 				    (lkg->sent->null_count != sent->length-1))
 				{
-					lgdebug(D_SLM, " (Extra, count > %zu)\n", lkg->sent->null_count);
+					lgdebug(D_SLM, " (Extra, count > %u)\n", lkg->sent->null_count);
 					match_found = false;
 					break;
 				}
@@ -390,7 +393,13 @@ bool sane_linkage_morphism(Sentence sent, Linkage lkg, Parse_Options opts)
 			break;
 		}
 
-		if (verbosity_level(D_SLM)) prt_error("%s", cdj->word_string);
+		if (verbosity_level(D_SLM))
+		{
+			if (cdj->is_category == 0)
+				prt_error("%s", cdj->word_string);
+			else
+				prt_error("Category[0]:%u", cdj->category[0].num);
+		}
 
 		match_found = false;
 		/* Proceed in all the paths in which the word is found. */
@@ -447,13 +456,13 @@ bool sane_linkage_morphism(Sentence sent, Linkage lkg, Parse_Options opts)
 	 * count when islands_ok=0. */
 	if (match_found)
 	{
-		size_t count_found =
+		unsigned int count_found =
 			opts->islands_ok ? num_islands(lkg, wpp->path) : null_count_found;
 
 		if ((count_found != lkg->sent->null_count) &&
 		    (lkg->sent->null_count != sent->length-1) && (count_found != sent->length))
 		{
-			lgdebug(D_SLM, "Null count mismatch: Found %zu != null_count %zu\n",
+			lgdebug(D_SLM, "Null count mismatch: Found %u != null_count %u\n",
 					  count_found, lkg->sent->null_count);
 			match_found = false;
 		}
