@@ -144,7 +144,7 @@ static void add_to_table_entry(unsigned int tsize, Match_node **table,
 	for (Match_node *m = sbin->head; NULL != m; m = m_next)
 	{
 		Connector *c = (0 == dir) ? m->d->left : m->d->right;
-		assert(NULL != c);
+		assert(NULL != c, "NULL connector");
 
 		Match_node **xl = get_match_table_entry(tsize, table, c, dir);
 		assert(&match_list_not_found != xl, "get_match_table_entry: Overflow");
@@ -225,6 +225,8 @@ static void sort_by_nearest_word(Match_node *m, sortbin *sbin, int nearest_word)
 
 fast_matcher_t* alloc_fast_matcher(const Sentence sent, unsigned int *ncu[])
 {
+	assert(sent->length > 0, "Sentence length is 0");
+
 	fast_matcher_t *ctxt;
 
 	ctxt = (fast_matcher_t *) xalloc(sizeof(fast_matcher_t));
@@ -239,13 +241,13 @@ fast_matcher_t* alloc_fast_matcher(const Sentence sent, unsigned int *ncu[])
 	ctxt->match_list = xalloc(ctxt->match_list_size * sizeof(*ctxt->match_list));
 	ctxt->match_list_end = 0;
 
-	if (NULL != sent->fm_Match_node)
+	if (NULL != sent->Match_node_pool)
 	{
-		pool_reuse(sent->fm_Match_node);
+		pool_reuse(sent->Match_node_pool);
 	}
 	else
 	{
-		sent->fm_Match_node =
+		sent->Match_node_pool =
 			pool_new(__func__, "Match_node",
 			         /*num_elements*/2048, sizeof(Match_node),
 			         /*zero_out*/false, /*align*/true, /*exact*/false);
@@ -254,7 +256,6 @@ fast_matcher_t* alloc_fast_matcher(const Sentence sent, unsigned int *ncu[])
 	sortbin *sbin = alloca(sent->length * sizeof(sortbin));
 
 	/* Calculate the sizes of the hash tables. */
-	size_t max_tsize = next_power_of_two_up(sent->dict->contable.num_uc);
 	unsigned int num_headers = 0;
 	Match_node **memblock_headers;
 	Match_node **hash_table_header;
@@ -273,7 +274,6 @@ fast_matcher_t* alloc_fast_matcher(const Sentence sent, unsigned int *ncu[])
 			else
 			{
 				tsize = next_power_of_two_up(3 * n); /* At least 66% free. */
-				tsize = MIN(max_tsize,  tsize);
 			}
 
 			ncu[dir][w] = tsize;
@@ -300,7 +300,7 @@ fast_matcher_t* alloc_fast_matcher(const Sentence sent, unsigned int *ncu[])
 		{
 			if (d->left != NULL)
 			{
-				Match_node *m = pool_alloc(sent->fm_Match_node);
+				Match_node *m = pool_alloc(sent->Match_node_pool);
 				m->d = d;
 				sort_by_nearest_word(m, sbin, d->left->nearest_word);
 			}
@@ -309,7 +309,7 @@ fast_matcher_t* alloc_fast_matcher(const Sentence sent, unsigned int *ncu[])
 		{
 			if (d->right != NULL)
 			{
-				Match_node *m = pool_alloc(sent->fm_Match_node);
+				Match_node *m = pool_alloc(sent->Match_node_pool);
 				m->d = d;
 				sort_by_nearest_word(m, sbin, d->right->nearest_word);
 			}
@@ -338,7 +338,8 @@ fast_matcher_t* alloc_fast_matcher(const Sentence sent, unsigned int *ncu[])
 		}
 	}
 
-	assert(memblock_headers + num_headers == hash_table_header);
+	assert(memblock_headers + num_headers == hash_table_header,
+	   "Mismatch header sizes");
 	return ctxt;
 }
 
@@ -469,12 +470,10 @@ typedef struct
  * (The first wordgraph word is used for cache validity indication,
  * but there is only one most of the times anyway.)
  */
-#define ALT_CONNECTION_POSSIBLE
 #define OPTIMIZE_EN
 static bool alt_connection_possible(Connector *c1, Connector *c2,
                                     gword_cache *c_con)
 {
-#ifdef ALT_CONNECTION_POSSIBLE
 	bool same_alternative = false;
 
 #ifdef OPTIMIZE_EN
@@ -507,9 +506,6 @@ static bool alt_connection_possible(Connector *c1, Connector *c2,
 	c_con->gword = c1->originating_gword->o_gword;
 
 	return same_alternative;
-#else
-	return true;
-#endif /* ALT_CONNECTION_POSSIBLE */
 }
 
 /**
@@ -559,22 +555,21 @@ form_match_list(fast_matcher_t *ctxt, int w,
 	size_t front = get_match_list_position(ctxt);
 	Match_node *ml = NULL, *mr = NULL; /* Initialize in case of NULL lc or rc. */
 	match_cache mc;
-	gword_cache gc;
-
-	gc.same_alternative = false;
+	gword_cache gc = { .same_alternative = false };
 
 	/* Get the lists of candidate matching disjuncts of word w for lc and
-	 * rc.  Consider each of these lists only if the length_limit of lc
-	 * rc and also w, is not greater then the distance between their word
-	 * and the word w. */
-	if ((lc != NULL) && ((w - lw) <= lc->length_limit))
+	 * rc. Consider each of these lists only if the farthest_word of lc/rc
+	 * reaches at least the word w.
+	 * Note: The commented out (w <= lc->farthest_word) is checked at the
+	 * callers and is left here for documentation. */
+	if ((lc != NULL) /* && (w <= lc->farthest_word) */)
 	{
 		ml = *get_match_table_entry(ctxt->l_table_size[w], ctxt->l_table[w], lc, 0);
 	}
 	if ((lc != NULL) && (ml == NULL)) /* lc optimization */
 		return terminate_match_list(ctxt, -1, front, w, lc, lw, rc, rw);
 
-	if ((rc != NULL) && ((rw - w) <= rc->length_limit))
+	if ((rc != NULL) && (w >= rc->farthest_word))
 	{
 		mr = *get_match_table_entry(ctxt->r_table_size[w], ctxt->r_table[w], rc, 1);
 	}
@@ -602,7 +597,7 @@ form_match_list(fast_matcher_t *ctxt, int w,
 	for (mx = ml; mx != NULL; mx = mx->next)
 	{
 		if (mx->d->left->nearest_word < lw) break;
-		if ((w - lw) > mx->d->left->length_limit) continue;
+		if (lw < mx->d->left->farthest_word) continue;
 
 		mx->d->match_left = do_match_with_cache(mx->d->left, lc, &mc) &&
 		                    alt_connection_possible(mx->d->left, lc, &gc);
@@ -627,7 +622,7 @@ form_match_list(fast_matcher_t *ctxt, int w,
 	gc.gword = NULL;
 	for (mx = mr; mx != mr_end; mx = mx->next)
 	{
-		if ((rw - w) > mx->d->right->length_limit) continue;
+		if (rw > mx->d->right->farthest_word) continue;
 
 		if ((lc != NULL) && !mx->d->match_left) continue; /* lc optimization */
 		mx->d->match_right = do_match_with_cache(mx->d->right, rc, &mc) &&
