@@ -36,6 +36,27 @@ static Linkage linkage_array_new(int num_to_alloc)
 	return lkgs;
 }
 
+static void find_unused_disjuncts(Sentence sent, extractor_t *pex)
+{
+	const size_t disjunct_used_sz =
+		sizeof(bool) * sent->wildcard_word_num_disjuncts;
+	sent->disjunct_used = malloc(disjunct_used_sz);
+
+	memset(sent->disjunct_used, 0, disjunct_used_sz);
+
+	if (pex != NULL)
+		mark_used_disjuncts(pex, sent->disjunct_used);
+
+	if (verbosity_level(+D_PARSE))
+	{
+		unsigned int num_unused = 0;
+		for (unsigned int i = 0; i < sent->wildcard_word_num_disjuncts; i++)
+			if (!sent->disjunct_used[i]) num_unused++;
+		prt_error("Info: Unused disjuncts %u/%u\n", num_unused,
+		          sent->wildcard_word_num_disjuncts);
+	}
+}
+
 static bool setup_linkages(Sentence sent, extractor_t* pex,
                           fast_matcher_t* mchxt,
                           count_context_t* ctxt,
@@ -44,7 +65,7 @@ static bool setup_linkages(Sentence sent, extractor_t* pex,
 	bool overflowed = build_parse_set(pex, sent, mchxt, ctxt, sent->null_count, opts);
 	print_time(opts, "Built parse set");
 
-	if (overflowed && (1 < opts->verbosity))
+	if (overflowed && (1 < opts->verbosity) && !IS_GENERATION(sent->dict))
 	{
 		err_ctxt ec = { sent };
 		err_msgc(&ec, lg_Warn, "Count overflow.\n"
@@ -90,11 +111,29 @@ static void print_chosen_disjuncts_words(const Linkage lkg, bool prt_optword)
 		const char *djw; /* disjunct word - the chosen word */
 
 		if (NULL == cdj)
+		{
 			djw = (prt_optword && lkg->sent->word[i].optional) ? "{}" : "[]";
-		else if ('\0' == cdj->word_string[0])
-			djw = "\\0"; /* null string - something is wrong */
+		}
+		else if (0 == cdj->is_category)
+		{
+			if ('\0' == cdj->word_string[0])
+				djw = "\\0"; /* null string - something is wrong */
+			else
+				djw = cdj->word_string;
+		}
 		else
-			djw = cdj->word_string;
+		{
+			if ((NULL == cdj->category))
+			{
+				djw = "\\0"; /* something is wrong */
+			}
+			else
+			{
+				char *cbuf = alloca(32); /* much more space than needed */
+				snprintf(cbuf, 32, "Category[0]:%u", cdj->category[0].num);
+				djw = cbuf;
+			}
+		}
 
 		dyn_strcat(djwbuf, djw);
 		dyn_strcat(djwbuf, " ");
@@ -103,6 +142,17 @@ static void print_chosen_disjuncts_words(const Linkage lkg, bool prt_optword)
 	dyn_str_delete(djwbuf);
 }
 
+/**
+ * Return \c true iff \p sent has an optional word.
+ */
+static bool optional_word_exists(Sentence sent)
+{
+	for (WordIdx w = 0; w < sent->length; w++)
+		if (sent->word[w].optional) return true;
+
+	return false;
+}
+
 #define D_PL 7
 /**
  * This fills the linkage array with morphologically-acceptable
@@ -150,6 +200,8 @@ static void process_linkages(Sentence sent, extractor_t* pex,
 		maxtries = sent->num_linkages_alloced;
 	}
 
+	bool need_sane_morphism = !IS_GENERATION(sent->dict) ||
+	                          optional_word_exists(sent);
 	bool need_init = true;
 	for (itry=0; itry<maxtries; itry++)
 	{
@@ -174,28 +226,36 @@ static void process_linkages(Sentence sent, extractor_t* pex,
 			print_chosen_disjuncts_words(lkg, /*prt_opt*/true);
 		}
 
-		if (sane_linkage_morphism(sent, lkg, opts))
+		if (need_sane_morphism)
 		{
-			remove_empty_words(lkg);
-
-			if (verbosity_level(+D_PL))
+			if (sane_linkage_morphism(sent, lkg, opts))
 			{
-				err_msg(lg_Debug, "chosen_disjuncts after:\n\\");
-				print_chosen_disjuncts_words(lkg, /*prt_opt*/false);
-			}
+				remove_empty_words(lkg);
 
-			need_init = true;
-			in++;
-			if (in >= sent->num_linkages_alloced) break;
-		}
-		else
-		{
-			N_invalid_morphism++;
-			lkg->num_links = 0;
-			lkg->num_words = sent->length;
-			// memset(lkg->link_array, 0, lkg->lasz * sizeof(Link));
-			memset(lkg->chosen_disjuncts, 0, sent->length * sizeof(Disjunct *));
+				if (verbosity_level(+D_PL))
+				{
+					err_msg(lg_Debug, "chosen_disjuncts after:\n\\");
+					print_chosen_disjuncts_words(lkg, /*prt_opt*/false);
+				}
+			}
+			else
+			{
+				N_invalid_morphism++;
+				lkg->num_links = 0;
+				lkg->num_words = sent->length;
+				// memset(lkg->link_array, 0, lkg->lasz * sizeof(Link));
+				memset(lkg->chosen_disjuncts, 0, sent->length * sizeof(Disjunct *));
+
+				continue;
+			}
 		}
+
+		if (IS_GENERATION(sent->dict))
+			compute_generated_words(sent, lkg);
+
+		need_init = true;
+		in++;
+		if (in >= sent->num_linkages_alloced) break;
 	}
 
 	/* The last one was alloced, but never actually used. Free it. */
@@ -229,6 +289,13 @@ static void sort_linkages(Sentence sent, Parse_Options opts)
 	print_time(opts, "Sorted all linkages");
 }
 
+static void notify_no_complete_linkages(unsigned int null_count,
+                                        unsigned int max_null_count)
+{
+		if ((0 == null_count) && (0 < max_null_count) && verbosity > 0)
+			prt_error("No complete linkages found.\n");
+}
+
 /**
  * classic_parse() -- parse the given sentence.
  * Perform parsing, using the original link-grammar parsing algorithm
@@ -304,10 +371,10 @@ void classic_parse(Sentence sent, Parse_Options opts)
 
 	/* Build lists of disjuncts */
 	prepare_to_parse(sent, opts);
-	if (resources_exhausted(opts->resources)) return;
+	if (resources_exhausted(opts->resources)) return; /* Nothing to free yet. */
 
 	Tracon_sharing *ts_pruning = pack_sentence_for_pruning(sent);
-	free_sentence_disjuncts(sent);
+	free_sentence_disjuncts(sent, /*category_too*/false);
 
 	if (one_step_parse)
 	{
@@ -323,6 +390,14 @@ void classic_parse(Sentence sent, Parse_Options opts)
 	{
 		sent->null_count = nl;
 
+		/* We may be here again for parsing with a higher null_count since
+		 * num_valid_linkages of the previous parse was 0 because all the
+		 * linkages had P.P. violations. Ensure that in case of a timeout we
+		 * will not end up with the previous num_linkages_found. */
+		sent->num_linkages_found = 0;
+		sent->num_valid_linkages = 0;
+		sent->num_linkages_post_processed = 0;
+
 		if (needed_prune_level > current_prune_level)
 		{
 			current_prune_level = needed_prune_level;
@@ -337,20 +412,21 @@ void classic_parse(Sentence sent, Parse_Options opts)
 			more_pruning_possible =
 				one_step_parse && (current_prune_level != MAX_SENTENCE);
 
-			unsigned int expexted_null_count =
+			unsigned int expected_null_count =
 				pp_and_power_prune(sent, ts_pruning, current_prune_level, opts,
 				                   ncu);
-			if (expexted_null_count > nl)
+			if (expected_null_count > nl)
 			{
 				if (opts->verbosity >= D_USER_TIMES)
 				{
 					prt_error("#### Skip parsing (w/%u ", nl);
-					if (expexted_null_count-1 > nl)
-						prt_error("to %u nulls)\n", expexted_null_count-1);
+					if (expected_null_count-1 > nl)
+						prt_error("to %u nulls)\n", expected_null_count-1);
 					else
 						prt_error("null%s)\n", (nl != 1) ? "s" : "");
 				}
-				nl = expexted_null_count-1;
+				notify_no_complete_linkages(nl, max_null_count);
+				nl = expected_null_count-1;
 				/* To get a result, parse w/null count which is at most one less
 				 * than the number of tokens (w/all nulls there is no linkage). */
 				if (nl == sent->length-1) nl--;
@@ -360,82 +436,100 @@ void classic_parse(Sentence sent, Parse_Options opts)
 
 		if (NULL != ts_pruning)
 		{
+
+			free_tracon_sharing(ts_parsing);
 			ts_parsing = pack_sentence_for_parsing(sent);
 			print_time(opts, "Encoded for parsing");
 
 			if (!more_pruning_possible)
 			{
 				/* At this point no further pruning will be done. Free the
-				 * pruning and parsing tracon stuff here instead of at the end. */
-				if (NULL != ts_pruning)
-				{
-					free(ts_pruning->memblock);
-					free_tracon_sharing(ts_pruning);
-					ts_pruning = NULL;
-					if (NULL != saved_memblock)
-						free(saved_memblock);
-				}
-
-				free_tracon_sharing(ts_parsing);
-				ts_parsing = NULL;
+				 * pruning tracon stuff here instead of at the end. */
+				free(ts_pruning->memblock);
+				free_tracon_sharing(ts_pruning);
+				ts_pruning = NULL;
+				if (NULL != saved_memblock)
+					free(saved_memblock);
 			}
 
 			gword_record_in_connector(sent);
-			if (resources_exhausted(opts->resources)) break;
-
-			free_count_context(ctxt, sent);
-			ctxt = alloc_count_context(sent);
 
 			free_fast_matcher(sent, mchxt);
 			mchxt = alloc_fast_matcher(sent, ncu);
 			print_time(opts, "Initialized fast matcher");
+			if (resources_exhausted(opts->resources)) goto parse_end_cleanup;
 		}
 
-		if (resources_exhausted(opts->resources)) break;
 		free_linkages(sent);
 
+		free_count_context(ctxt, sent);
+		ctxt = alloc_count_context(sent, ts_parsing);
+
 		sent->num_linkages_found = do_parse(sent, mchxt, ctxt, opts);
 
 		print_time(opts, "Counted parses (%d w/%u null%s)",
 		           sent->num_linkages_found, sent->null_count,
 		           (sent->null_count != 1) ? "s" : "");
 
-		extractor_t * pex = extractor_new(sent->length, sent->rand_state);
-		bool ovfl = setup_linkages(sent, pex, mchxt, ctxt, opts);
-		process_linkages(sent, pex, ovfl, opts);
-		free_extractor(pex);
-
-		post_process_lkgs(sent, opts);
-
-		if (sent->num_valid_linkages > 0) break;
-
-		if (verbosity >= D_USER_INFO)
+		/* In case of a timeout, the linkage is partial and may be
+		 * inconsistent. It is also usually different on each run.
+		 * So in that case, pretend that the linkage count is 0. */
+		if (resources_exhausted(opts->resources))
 		{
-			if ((sent->num_valid_linkages == 0) &&
-			    (sent->num_linkages_post_processed > 0) &&
-			    ((int)opts->linkage_limit < sent->num_linkages_found))
-				prt_error("Info: All examined linkages (%zu) had P.P. violations.\n"
-				          "Consider increasing the linkage limit.\n"
-				          "At the command line, use !limit\n",
-				          sent->num_linkages_post_processed);
+			sent->num_linkages_found = 0;
+			goto parse_end_cleanup;
 		}
 
-		if ((0 == nl) && (0 < max_null_count) && verbosity > 0)
-			prt_error("No complete linkages found.\n");
+		if (sent->num_linkages_found > 0)
+		{
+			extractor_t * pex =
+				extractor_new(sent->length, sent->rand_state, IS_GENERATION(sent->dict));
+			bool ovfl = setup_linkages(sent, pex, mchxt, ctxt, opts);
+			process_linkages(sent, pex, ovfl, opts);
+			if (IS_GENERATION(sent->dict))
+			    find_unused_disjuncts(sent, pex);
+			free_extractor(pex);
 
-		free_tracon_sharing(ts_parsing);
-		ts_parsing = NULL;
+			post_process_lkgs(sent, opts);
+			if (resources_exhausted(opts->resources))
+			{
+				sent->num_linkages_found = 0;
+				sent->num_valid_linkages = 0;
+				sent->num_linkages_post_processed = 0;
+				goto parse_end_cleanup;
+			}
+
+			if (sent->num_valid_linkages > 0) break;
+
+			if (verbosity >= D_USER_INFO)
+			{
+				if ((sent->num_valid_linkages == 0) &&
+				    (sent->num_linkages_post_processed > 0) &&
+				    ((int)opts->linkage_limit < sent->num_linkages_found) &&
+				    (PARSE_NUM_OVERFLOW >= sent->num_linkages_found))
+					prt_error("Info: All examined linkages (%zu) had P.P. violations.\n"
+					          "Consider increasing the linkage limit.\n"
+					          "At the command line, use !limit\n",
+					          sent->num_linkages_post_processed);
+			}
+		}
+
+		notify_no_complete_linkages(nl, max_null_count);
 	}
+	if ((sent->num_linkages_found == 0) && IS_GENERATION(sent->dict))
+		find_unused_disjuncts(sent, NULL);
+
 	sort_linkages(sent, opts);
 
+parse_end_cleanup:
 	if (NULL != ts_pruning)
 	{
+		free_categories(sent);
 		free(ts_pruning->memblock);
 		free_tracon_sharing(ts_pruning);
 		free(saved_memblock);
 	}
 	free_tracon_sharing(ts_parsing);
-
 	free_count_context(ctxt, sent);
 	free_fast_matcher(sent, mchxt);
 }
