@@ -27,7 +27,7 @@
 #include "tokenize/word-structures.h"   // Word_struct
 #include "tokenize/wordgraph.h"
 
-#define D_PRUNE 5      /* Debug level for this file. */
+#define D_PRUNE 5      /* Debug level for this file (6 for pp_prune()) */
 
 /* To debug pp_prune(), touch this file, run "make CPPFLAGS=-DDEBUG_PP_PRUNE",
  * and then run: link-parser -v=5 -debug=prune.c . */
@@ -36,7 +36,7 @@
 #endif
 
 #ifdef DEBUG_PP_PRUNE
-#define ppdebug(...) lgdebug(+D_PRUNE, __VA_ARGS__)
+#define ppdebug(...) lgdebug(+D_PRUNE+1, __VA_ARGS__)
 #else
 #define ppdebug(...)
 #endif
@@ -49,9 +49,10 @@
 
 typedef uint8_t WordIdx_m;     /* Storage representation of word index */
 
+/* The dimension of the 2-element arrays below (unless stated otherwise)
+ * is used as follows: [0] - left side; [1] - right side. */
+
 /* Per-word minimum/maximum link distance descriptor.
- * The dimension of the 2-element arrays below is used as follows:
- * [0] - left side; [1] - right side.
  *
  * nw is the minimum nearest_word of the shallow connectors.
  * fw is the maximum farthest_word of the shallow connectors.
@@ -89,7 +90,7 @@ typedef struct
 {
 	WordIdx_m nw[2];         /* minimum link distance */
 	WordIdx_m nw_perjet[2];  /* the same, ignoring missing jets */
-	WordIdx_m nw_unidir[2];  /* the same, but only for unidirectional jets */
+	WordIdx_m nw_unidir[2];  /* the same, but for unidirectional disjuncts */
 	WordIdx_m fw[2];         /* maximum link distance */
 } mlink_table;
 
@@ -104,10 +105,8 @@ typedef struct power_table_s power_table;
 struct power_table_s
 {
 	unsigned int power_table_size;
-	unsigned int *l_table_size;  /* the sizes of the hash tables */
-	unsigned int *r_table_size;
-	C_list *** l_table;
-	C_list *** r_table;
+	unsigned int *table_size[2];  /* the sizes of the hash tables */
+	C_list ***table[2];
 	Pool_desc *memory_pool;
 };
 
@@ -141,6 +140,42 @@ struct prune_context_s
 	int N_xlink;      /* counts linked interval link-crossing rejections */
 };
 
+#ifdef DEBUG
+GNUC_UNUSED static void print_power_table_entry(power_table *pt, int w, int dir)
+{
+	C_list **t = pt->table[w][dir];
+	unsigned int size = pt->table_size[w][dir];
+
+	if (size == 1) return;
+	printf("w%d dir%d size=%u:\n", w, dir, size);
+
+	for (unsigned int i = 0; i < size; i++)
+	{
+		if (t[i] == NULL) continue;
+
+		printf(" [%u]: ", i);
+		for (C_list *cl = t[i]; cl != NULL; cl = cl->next)
+		{
+			char *cstr = print_one_connector_str(cl->c, "lrs");
+			printf("%s", cstr);
+			free(cstr);
+			if (cl->next != NULL) printf(" ");
+		}
+		printf("\n");
+	}
+}
+
+GNUC_UNUSED static void print_power_table(Sentence sent, power_table *pt)
+{
+	printf("power table:\n");
+	for (WordIdx w = 0; w < sent->length; w++)
+	{
+		for (int dir = 0; dir < 2; dir++)
+			print_power_table_entry(pt, w, dir);
+	}
+}
+#endif
+
 /*
  * Here is what you've been waiting for: POWER-PRUNE
  *
@@ -150,8 +185,8 @@ struct prune_context_s
  *     nearer and nearer words.
  *
  *  2) Two deep connectors cannot attach to each other; i.e. a deep
- *     connector can only attache to a shallow one, and a shallow
- *     connector can attache to any connector.
+ *     connector can only attach to a shallow one, and a shallow
+ *     connector can attach to any connector.
  *     (A connectors is deep if it is not the first in its list; it
  *     is shallow if it is the first in its list; it is deepest if it
  *     is the last on its list.)
@@ -205,13 +240,9 @@ struct prune_context_s
 static void power_table_delete(power_table *pt)
 {
 	pool_delete(pt->memory_pool);
-	for (WordIdx w = 0; w < pt->power_table_size; w++)
-	{
-		xfree((char *)pt->l_table[w], pt->l_table_size[w] * sizeof (C_list *));
-		xfree((char *)pt->r_table[w], pt->r_table_size[w] * sizeof (C_list *));
-	}
-	xfree(pt->l_table_size, 2 * pt->power_table_size * sizeof(unsigned int));
-	xfree(pt->l_table, 2 * pt->power_table_size * sizeof(C_list **));
+	free(pt->table_size[0]);
+	free(pt->table[0][0]);
+	free(pt->table[0]);
 }
 
 static C_list **get_power_table_entry(unsigned int size, C_list **t,
@@ -257,10 +288,10 @@ static void put_into_power_table(Pool_desc *mp, unsigned int size, C_list **t,
 static void power_table_alloc(Sentence sent, power_table *pt)
 {
 	pt->power_table_size = sent->length;
-	pt->l_table_size = xalloc (2 * sent->length * sizeof(unsigned int));
-	pt->r_table_size = pt->l_table_size + sent->length;
-	pt->l_table = xalloc (2 * sent->length * sizeof(C_list **));
-	pt->r_table = pt->l_table + sent->length;
+	pt->table_size[0] = malloc (2 * sent->length * sizeof(unsigned int));
+	pt->table_size[1] = pt->table_size[0] + sent->length;
+	pt->table[0] = malloc (2 * sent->length * sizeof(C_list **));
+	pt->table[1] = pt->table[0] + sent->length;
 }
 
 /**
@@ -289,9 +320,6 @@ static void power_table_alloc(Sentence sent, power_table *pt)
 static void power_table_init(Sentence sent, Tracon_sharing *ts, power_table *pt)
 {
 	Tracon_list *tl = ts->tracon_list;
-	unsigned int i;
-#define TOPSZ 32768
-	size_t lr_table_max_usage = MIN(sent->dict->contable.num_con, TOPSZ);
 
 	power_table_alloc(sent, pt);
 
@@ -299,112 +327,97 @@ static void power_table_init(Sentence sent, Tracon_sharing *ts, power_table *pt)
 	                   /*num_elements*/2048, sizeof(C_list),
 	                   /*zero_out*/false, /*align*/false, /*exact*/false);
 
+	/* The below uses variable-sized hash tables. The sizes are provided
+	 * in num_cnctrs_per_word[][], which is the number of different
+	 * uppercase connectors, indexed by direction and word. It is computed in
+	 * pack_connectors(). */
+
+	/* Calculate the sizes of the hash tables. */
+	unsigned int num_headers = 0;
+	C_list **memblock_headers;
+	C_list **hash_table_header;
+
+	unsigned int *ncu[2];
+	ncu[0] = alloca(sent->length * sizeof(*ncu[0]));
+	ncu[1] = alloca(sent->length * sizeof(*ncu[1]));
+
 	for (WordIdx w = 0; w < sent->length; w++)
 	{
-		size_t l_size, r_size;
-		C_list **l_t, **r_t;
-		size_t len;
-
-		/* The below uses variable-sized hash tables. This seems to
-		 * provide performance that is equal or better than the best
-		 * fixed-size performance.
-		 * The best fixed-size performance seems to come at about
-		 * a 1K table size, for both English and Russian. (Both have
-		 * about 100 fixed link-types, and many thousands of auto-genned
-		 * link types (IDxxx idioms for both, LLxxx suffix links for
-		 * Russian).  Pluses and minuses:
-		 * + small fixed tables are faster to initialize.
-		 * - small fixed tables have more collisions
-		 * - variable-size tables require counting connectors.
-		 *   (and the more complex code to go with)
-		 * CPU cache-size effects ...
-		 */
-		if (NULL != tl)
-			len = tl->num_cnctrs_per_word[0][w];
-		else
-			len = left_connector_count(sent->word[w].d);
-		len++; /* Ensure at least one empty entry for get_power_table_entry(). */
-		l_size = next_power_of_two_up(MIN(len, lr_table_max_usage));
-		pt->l_table_size[w] = l_size;
-		l_t = pt->l_table[w] = (C_list **) xalloc(l_size * sizeof(C_list *));
-		for (i=0; i<l_size; i++) l_t[i] = NULL;
-
-		if (NULL != tl)
-			len = tl->num_cnctrs_per_word[1][w];
-		else
-			len = right_connector_count(sent->word[w].d);
-		len++;
-		r_size = next_power_of_two_up(MIN(len, lr_table_max_usage));
-		pt->r_table_size[w] = r_size;
-		r_t = pt->r_table[w] = (C_list **) xalloc(r_size * sizeof(C_list *));
-		for (i=0; i<r_size; i++) r_t[i] = NULL;
-
-		if (NULL == tl)
+		for (size_t dir = 0; dir < 2; dir++)
 		{
-			/* Insert the deep connectors. */
-			for (Disjunct *d = sent->word[w].d; d != NULL; d = d->next)
+			unsigned int tsize;
+			unsigned int n = ts->num_cnctrs_per_word[dir][w];
+
+			if (0 == n)
 			{
-				Connector *c;
-
-				c = d->right;
-				if (c != NULL)
-				{
-					for (c = c->next; c != NULL; c = c->next)
-					{
-						put_into_power_table(mp, r_size, r_t, c);
-					}
-				}
-
-				c = d->left;
-				if (c != NULL)
-				{
-					for (c = c->next; c != NULL; c = c->next)
-					{
-						put_into_power_table(mp, l_size, l_t, c);
-					}
-				}
+				tsize = 1; /* Avoid parse-time table size checks. */
+			}
+			else
+			{
+				tsize = next_power_of_two_up(3 * n); /* At least 66% free. */
 			}
 
-			/* Insert the shallow connectors. */
-			for (Disjunct *d = sent->word[w].d; d != NULL; d = d->next)
+			ncu[dir][w] = tsize;
+			num_headers += tsize;
+		}
+	}
+
+	memblock_headers = malloc(num_headers * sizeof(C_list *));
+	memset(memblock_headers, 0, num_headers * sizeof(C_list *));
+	hash_table_header = memblock_headers;
+
+	for (WordIdx w = 0; w < sent->length; w++)
+	{
+		for (size_t dir = 0; dir < 2; dir++)
+		{
+			unsigned int tsize = ncu[dir][w];
+			C_list **t = hash_table_header;
+
+			hash_table_header += tsize;
+
+			pt->table[dir][w] = t;
+			pt->table_size[dir][w] = tsize;
+			memset(t, 0, sizeof(C_list *) * tsize);
+
+			if (NULL == tl)
 			{
 				Connector *c;
 
-				c = d->right;
-				if (c != NULL)
+				for (Disjunct *d = sent->word[w].d; d != NULL; d = d->next)
 				{
-					put_into_power_table(mp, r_size, r_t, c);
+					/* Insert the deep connectors. */
+					c = (dir == 0) ? d->left : d->right;
+					if (c == NULL) continue;
+
+					for (c = c->next; c != NULL; c = c->next)
+						put_into_power_table(mp, tsize, t, c);
 				}
-				c = d->left;
-				if (c != NULL)
+
+				for (Disjunct *d = sent->word[w].d; d != NULL; d = d->next)
 				{
-					put_into_power_table(mp, l_size, l_t, c);
+					/* Insert the shallow connectors. */
+					c = (dir == 0) ? d->left : d->right;
+					if (c == NULL) continue;
+
+					put_into_power_table(mp, tsize, t, c);
 				}
 			}
 		}
 	}
 
+	assert(memblock_headers + num_headers == hash_table_header,
+	   "Mismatch header sizes");
+
 	if (NULL != tl)
 	{
 		/* Bulk insertion with reference count. */
 
-		for (int dir = 0; dir < 2; dir++)
+		for (size_t dir = 0; dir < 2; dir++)
 		{
-			C_list ***tp;
-			unsigned int *sizep;
+			C_list ***tp = pt->table[dir];
+			unsigned int *sizep = pt->table_size[dir];
 			unsigned int sid_entries = tl->entries[dir];
 
-			if (dir == 0)
-			{
-				tp = pt->l_table;
-				sizep = pt->l_table_size;
-			}
-			else
-			{
-				tp = pt->r_table;
-				sizep = pt->r_table_size;
-			}
-
 			/* Insert the deep connectors, then the shallow ones. */
 			for (int shallow = 0; shallow < 2; shallow++)
 			{
@@ -434,9 +447,11 @@ static void power_table_init(Sentence sent, Tracon_sharing *ts, power_table *pt)
 static void clean_table(unsigned int size, C_list **t)
 {
 	/* Table entry tombstone. */
+#define UC_NUM_TOMBSTONE ((connector_hash_t)-1)
 	static condesc_t desc_no_match =
 	{
-		.uc_num = (connector_hash_t)-1, /* get_power_table_entry() will skip. */
+		.string = "TOMBSTONE",
+		.uc_num = UC_NUM_TOMBSTONE, /* get_power_table_entry() will skip. */
 	};
 	static Connector con_no_match =
 	{
@@ -561,27 +576,22 @@ static bool is_cross_mlink(prune_context *pc,
 		if (sent->word[w].optional) continue;
 		if (pc->is_null_word[w]) continue;
 
-		if ((w == lword+1) && (pc->ml[w].nw_unidir[1] > rword) &&
+		if ((w == lword+1) && (pc->ml[w].nw_perjet[1] > rword) &&
 			 !is_match(pc, left_table_search, lword, lc, w))
 		{
 			PR(L);
 			goto null_word_found;
 		}
-		if ((w == rword-1) && (pc->ml[w].nw_unidir[0] < lword) &&
+		if ((w == rword-1) && (pc->ml[w].nw_perjet[0] < lword) &&
 		    !is_match(pc, right_table_search, rword, rc, w))
 		{
 			PR(R);
 			goto null_word_found;
 		}
 
-		if ((pc->ml[w].nw_perjet[0] < lword) && (pc->ml[w].nw_unidir[1] > rword))
+		if ((pc->ml[w].nw_perjet[0] < lword) && (pc->ml[w].nw_perjet[1] > rword))
 		{
-			PR(L);
-			goto null_word_found;
-		}
-		if ((pc->ml[w].nw_unidir[0] < lword) && (pc->ml[w].nw_perjet[1] > rword))
-		{
-			PR(R);
+			PR(P);
 			goto null_word_found;
 		}
 
@@ -779,8 +789,8 @@ right_table_search(prune_context *pc, int w, Connector *c,
                    bool shallow, int word_c)
 {
 	power_table *pt = pc->pt;
-	unsigned int size = pt->r_table_size[w];
-	C_list **e = get_power_table_entry(size, pt->r_table[w], c);
+	unsigned int size = pt->table_size[1][w];
+	C_list **e = get_power_table_entry(size, pt->table[1][w], c);
 
 	for (C_list *cl = *e; cl != NULL; cl = cl->next)
 	{
@@ -795,7 +805,7 @@ right_table_search(prune_context *pc, int w, Connector *c,
 }
 
 /**
- * This returns TRUE if the right table of word w contains
+ * This returns TRUE if the left table of word w contains
  * a connector that can match to c.  shallow tells if c is shallow
  */
 static bool
@@ -803,8 +813,8 @@ left_table_search(prune_context *pc, int w, Connector *c,
                   bool shallow, int word_c)
 {
 	power_table *pt = pc->pt;
-	unsigned int size = pt->l_table_size[w];
-	C_list **e = get_power_table_entry(size, pt->l_table[w], c);
+	unsigned int size = pt->table_size[0][w];
+	C_list **e = get_power_table_entry(size, pt->table[0][w], c);
 
 	for (C_list *cl = *e; cl != NULL; cl = cl->next)
 	{
@@ -1081,6 +1091,7 @@ static int power_prune(Sentence sent, prune_context *pc, Parse_Options opts)
 
 					/* Discard the current disjunct. */
 					*dd = d->next; /* NEXT - set current disjunct to the next one */
+					if (d->is_category != 0) free(d->category);
 					pc->N_deleted[(int)bad]++;
 					continue;
 				}
@@ -1090,7 +1101,7 @@ static int power_prune(Sentence sent, prune_context *pc, Parse_Options opts)
 			}
 
 			if (check_null_word(pc, w)) extra_null_word = true;
-			clean_table(pt->r_table_size[w], pt->r_table[w]);
+			clean_table(pt->table_size[1][w], pt->table[1][w]);
 		}
 
 		if (pruning_pass_end(pc, "l->r", &total_deleted)) break;
@@ -1115,6 +1126,7 @@ static int power_prune(Sentence sent, prune_context *pc, Parse_Options opts)
 
 					/* Discard the current disjunct. */
 					*dd = d->next; /* NEXT - set current disjunct to the next one */
+					if (d->is_category != 0) free(d->category);
 					pc->N_deleted[(int)bad]++;
 					continue;
 				}
@@ -1124,13 +1136,13 @@ static int power_prune(Sentence sent, prune_context *pc, Parse_Options opts)
 			}
 
 			if (check_null_word(pc, w)) extra_null_word = true;
-			clean_table(pt->l_table_size[w], pt->l_table[w]);
+			clean_table(pt->table_size[0][w], pt->table[0][w]);
 		}
 
 		if (pruning_pass_end(pc, "r->l", &total_deleted)) break;
 
-		/* The verbose debug printouts revealed that the xlink counter is
-		 * never set after the first 2 passes. So neutralize the mlink table
+		/* The above debug printouts revealed that the xlink counter doesn't
+		 * get increased after the first 2 passes. So neutralize the mlink table
 		 * here to save a slight overhead. */
 		pc->ml = NULL;
 	}
@@ -1140,7 +1152,7 @@ static int power_prune(Sentence sent, prune_context *pc, Parse_Options opts)
 	if ((verbosity >= D_USER_TIMES) && !extra_null_word && (pc->null_words > 0))
 		snprintf(found_nulls, sizeof(found_nulls), ", found %u", pc->null_words);
 	print_time(opts, "power pruned (for %u null%s%s%s)",
-	           pc->null_links, (pc->null_links != 1) ? "s" : "",
+					  pc->null_links, (pc->null_links != 1) ? "s" : "",
 	           extra_null_word ? ", extra null" : "", found_nulls);
 	if (verbosity_level(D_PRUNE))
 	{
@@ -1296,8 +1308,8 @@ static void cms_table_delete(multiset_table *mt)
 static unsigned int cms_hash(const char *s)
 {
 	unsigned int i = 5381;
-	if (islower((int) *s)) s++; /* skip head-dependent indicator */
-	while (isupper((int) *s))
+	if (islower((unsigned int)*s)) s++; /* skip head-dependent indicator */
+	while (is_connector_name_char(*s))
 	{
 		i = ((i << 5) + i) + *s;
 		s++;
@@ -1335,13 +1347,13 @@ static void reset_last_criterion(multiset_table *cmt, const char *ctiterion)
 static bool can_form_link(const char *s, const char *t, const char *e)
 {
 	if (islower(*t)) t++; /* Skip head-dependent indicator */
-	while (isupper(*s))
+	while (is_connector_name_char(*s))
 	{
 		if (*s != *t) return false;
 		s++;
 		t++;
 	}
-	if (isupper(*t)) return false;
+	if (is_connector_name_char(*t)) return false;
 	while (*t != '\0')
 	{
 		if (*s == '\0') return true;
@@ -1480,7 +1492,7 @@ static bool all_connectors_exist(multiset_table *cmt, const char *pp_link)
 	ppdebug("check PP-link=%s\n", pp_link);
 
 	const char *s;
-	for (s = pp_link; isupper(*s); s++) {}
+	for (s = pp_link; is_connector_name_char(*s); s++) {}
 
 	/* The first iteration of the next loop handles the special case
 	 * which occurs if there were 0 subscripts. */
@@ -1495,7 +1507,7 @@ static bool all_connectors_exist(multiset_table *cmt, const char *pp_link)
 	return true;
 }
 
-static bool connecor_has_direction(Cms *cms, int dir)
+static bool connector_has_direction(Cms *cms, int dir)
 {
 	return ((dir == 0) && cms->left) || ((dir == 1) && cms->right);
 }
@@ -1511,12 +1523,12 @@ static bool any_possible_connection(multiset_table *cmt, const char *criterion)
 		ppdebug("TRY %s%s\n", connector_string(cms1->c), connector_signs(cms1));
 		for (int dir = 0; dir < 2; dir++)
 		{
-			if (!connecor_has_direction(cms1, dir)) continue;
+			if (!connector_has_direction(cms1, dir)) continue;
 			Connector *c = cms1->c;
 
 			for (Cms *cms2 = cmt->cms_table[h]; cms2 != NULL; cms2 = cms2->next)
 			{
-				if (!connecor_has_direction(cms2, 1-dir)) continue;
+				if (!connector_has_direction(cms2, 1-dir)) continue;
 				Connector *cfl = cms2->c;
 
 				if (easy_match_desc(cfl->desc, c->desc))
@@ -1749,7 +1761,7 @@ static int pp_prune(Sentence sent, Tracon_sharing *ts, Parse_Options opts)
 		}
 	}
 
-	lgdebug(+D_PRUNE, "Deleted %d (%d connector names)\n",
+	lgdebug(+D_PRUNE+1, "Deleted %d (%d connector names)\n",
 	        D_deleted, Cname_deleted);
 
 	cms_table_delete(cmt);
@@ -1773,21 +1785,10 @@ static void get_num_con_uc(Sentence sent,power_table *pt,
 	/* For each table, count the number of non-empty slots. */
 	for (WordIdx w = 0; w < sent->length; w++)
 	{
-		for (int dir = 0; dir < 2; dir++)
+		for (size_t dir = 0; dir < 2; dir++)
 		{
-			C_list **t;
-			unsigned int size;
-
-			if (dir == 0)
-			{
-				t = pt->l_table[w];
-				size = pt->l_table_size[w];
-			}
-			else
-			{
-				t = pt->r_table[w];
-				size = pt->r_table_size[w];
-			}
+			C_list **t = pt->table[dir][w];
+			unsigned int size = pt->table_size[dir][w];
 
 			unsigned int count = 0;
 			for (unsigned int h = 0; h < size; h++)
@@ -2213,9 +2214,8 @@ unsigned int pp_and_power_prune(Sentence sent, Tracon_sharing *ts,
 		}
 	}
 
-	/* No benefit for now to make additional pp_prune() & power_prune() -
-	 * additional deletions are very rare and even then most of the
-	 * times only one disjunct is deleted. */
+	/* It is not cost-effective to make additional pp_prune() & power_prune()
+	 * as most of the times too few disjunct are deleted. */
 
 	/* Initialize tentative values. */
 	unsigned int min_nulls = sent->null_count;
