@@ -15,9 +15,8 @@
 #include <string.h>
 
 #include "api-structures.h"
-#include "dict-common/dict-defines.h"    // For RIGHT_WALL_WORD
-#include "dict-common/dict-common.h"     // For Dictionary_s
-#include "dict-common/dict-defines.h"    // For MAX_WORD
+#include "dict-common/dict-common.h"    // Dictionary_s
+#include "dict-common/dict-defines.h"   // RIGHT_WALL_WORD, MAX_WORD
 #include "error.h"
 #include "linkage/linkage.h"
 #include "post-process/post-process.h"
@@ -77,7 +76,7 @@ struct CNode_s
  */
 static bool uppercompare(const char * s, const char * t)
 {
-	while (isupper(*s) || isupper(*t))
+	while (is_connector_name_char(*s) || is_connector_name_char(*t))
 	{
 		if (*s++ != *t++) return false;
 	}
@@ -616,7 +615,7 @@ static int add_constituent(con_context_t *ctxt, int c, const Linkage linkage,
 	if (l < 1) l=1;
 	if (r > nwords) r = nwords;
 	if (l > nwords) l = nwords;
-	assert(l <= r, "negative constituent length!" );
+	assert(l <= r, "negative constituent length!");
 
 	ctxt->constituent[c].type = string_set_add(name, ctxt->phrase_ss);
 	ctxt->constituent[c].left = l;
@@ -1325,7 +1324,7 @@ char * linkage_print_constituent_tree(Linkage linkage, ConstituentDisplayStyle m
 	{
 		return print_flat_constituents(linkage);
 	}
-	prt_error("Warning: Illegal mode %d for printing constituents\n"
+	prt_error("Warning: Illegal mode %u for printing constituents\n"
 	          "Allowed values: %d to %d\n", mode, NO_DISPLAY, MAX_STYLES);
 	return NULL;
 }
