@@ -24,8 +24,8 @@
 #endif
 #include <signal.h>    /* SIG* */
 
-#include <print/print-util.h> /* for append_string */
-#include <utilities.h> /* for dyn_str functions and UNREACHABLE */
+#include "print/print-util.h" /* for append_string */
+#include "utilities.h" /* for dyn_str functions and UNREACHABLE */
 #endif /* USE_WORDGRAPH_DISPLAY */
 
 #include "api-structures.h"
@@ -93,7 +93,7 @@ GNUC_UNUSED const char *gword_morpheme(Sentence sent, const Gword *w)
 			break;
 		default:
 			/* No truncation is expected. */
-			snprintf(buff, sizeof(buff), "MT_%d", w->morpheme_type);
+			snprintf(buff, sizeof(buff), "MT_%d", (int)w->morpheme_type);
 			mt = string_set_add(buff, sent->string_set);
 	}
 
@@ -646,7 +646,7 @@ finish:
 #else
 bool sentence_display_wordgraph(Sentence sent, const char *modestr)
 {
-		prt_error("Error: Library not configured with wordgraph-display\n");
-		return false;
+	prt_error("Error: Library not configured with wordgraph-display\n");
+	return false;
 }
 #endif /* USE_WORDGRAPH_DISPLAY */
