@@ -33,6 +33,8 @@
 	#include <windows.h>
 #endif /* _WIN32 */
 
+#include "link-includes.h"
+#include "error.h"
 #include "utilities.h"
 
 /* This file contains general utilities that fix, enhance OS-provided
@@ -685,7 +687,7 @@ win32_getlocale (void)
 char * get_default_locale(void)
 {
 	const char *lc_vars[] = {"LC_ALL", "LC_CTYPE", "LANG", NULL};
-	char *ev;
+	char *ev = NULL;
 	const char **evname;
 	char *locale = NULL;
 
@@ -757,7 +759,7 @@ bool strtodC(const char *s, float *r)
 {
 	char *err;
 
-#ifdef HAVE_LOCALE_T
+#if defined(HAVE_LOCALE_T) && !defined(__sun__)
 	double val = strtod_l(s, &err, get_C_LC_NUMERIC());
 #else
 	/* dictionary_setup_locale() invokes setlocale(LC_NUMERIC, "C") */
@@ -766,7 +768,7 @@ bool strtodC(const char *s, float *r)
 
 	if ('\0' != *err) return false; /* *r unaffected */
 
-	*r = val;
+	*r = (float)val;
 	return true;
 }
 
