@@ -28,8 +28,7 @@
 #include <xlocale.h>
 #endif /* HAVE_LOCALE_T_IN_XLOCALE_H */
 
-#include "error.h"
-#include "lg_assert.h"
+#include "link-includes.h"
 
 #ifdef HAVE_ALLOCA_H
 # include <alloca.h>
@@ -253,8 +252,9 @@ static inline char *_strndupa3(char *new_s, const char *s, size_t n)
 #if __GNUC__
 #define GCC_DIAGNOSTIC
 #define UNREACHABLE(x) (__extension__ ({if (x) __builtin_unreachable();}))
-#define GNUC_MALLOC __attribute__ ((malloc))
-#define GNUC_UNUSED __attribute__ ((unused))
+#define GNUC_MALLOC __attribute__ ((__malloc__))
+#define GNUC_UNUSED __attribute__ ((__unused__))
+#define GNUC_NORETURN __attribute__ ((__noreturn__))
 #define NO_SAN __attribute__ ((no_sanitize_address, no_sanitize_undefined))
 
 /* Define when configuring with ASAN/UBSAN - for fast dict load (of course
@@ -266,11 +266,20 @@ static inline char *_strndupa3(char *new_s, const char *s, size_t n)
 #define NO_SAN_DICT
 #endif
 
+#ifndef DONT_EXPECT
+#define likely(x)      __builtin_expect(!!(x), 1)
+#define unlikely(x)    __builtin_expect(!!(x), 0)
+#endif
+
 #else
 #define UNREACHABLE(x)
 #define GNUC_MALLOC
 #define GNUC_UNUSED
+#define GNUC_NORETURN
 #define NO_SAN_DICT
+
+#define likely(x)
+#define unlikely(x)
 #endif
 
 
